# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import dataclasses
import datetime

import sqlalchemy as sa
from nbxmpp import JID
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Select
from sqlalchemy import select
from sqlalchemy import types
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import relationship
from sqlalchemy.sql import expression as expr
from sqlalchemy.types import TypeEngine

from gajim.common import app
from gajim.common.storage.base import EpochTimestampType
from gajim.common.storage.base import JIDType
from gajim.common.storage.base import JSONType
from gajim.common.storage.base import StrValueMissingType
from gajim.common.storage.base import VALUE_MISSING
from gajim.common.storage.base import ValueMissingT


class Base(DeclarativeBase):
    type_annotation_map: dict[type[Any], TypeEngine[Any]] = {
        float: types.REAL(),
        str: types.TEXT(),
        int: types.INTEGER(),
    }


class UtilMixin:
    def get_upsert_values(self) -> dict[str, str]:
        values = {
            f.name: getattr(self, f.name)
            for f in dataclasses.fields(self)  # pyright: ignore
        }
        for key, value in list(values.items()):
            if value is VALUE_MISSING:
                values.pop(key)
                continue

            if key not in self.__upsert_cols__:  # pyright: ignore
                values.pop(key)
        return values

    def get_insert_values(self):
        values = {
            f.name: getattr(self, f.name)
            for f in dataclasses.fields(self)  # pyright: ignore
        }
        for key, value in list(values.items()):
            if value is VALUE_MISSING:
                values.pop(key)
                continue

            if key in self.__no_table_cols__:  # pyright: ignore
                values.pop(key)
        return values


class Account(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'account'
    __table_args__ = (Index('idx_account', 'jid', unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    jid: Mapped[JID] = mapped_column(JIDType)


class Remote(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'remote'
    __table_args__ = (Index('idx_remote', 'jid', unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    jid: Mapped[JID] = mapped_column(JIDType)


class Thread(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'thread'
    __index_cols__ = ['id', 'fk_remote_pk', 'fk_account_pk']
    __no_table_cols__ = []
    __table_args__ = (Index('idx_thread', *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE')
    )
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'))

    id: Mapped[str]

    def get_select_stmt(self) -> Select[Any]:
        return select(Thread).where(
            Thread.id == self.id,
            Thread.fk_remote_pk == self.fk_remote_pk,
            Thread.fk_account_pk == self.fk_account_pk,
        )


class Occupant(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'occupant'
    __index_cols__ = ['id', 'fk_remote_pk', 'fk_account_pk']
    __upsert_cols__ = ['fk_real_remote_pk', 'nickname', 'avatar_sha', 'updated_at']
    __no_table_cols__ = ['account_', 'remote_jid_', 'real_remote_jid_', 'real_remote']
    __table_args__ = (Index('idx_occupant', *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    id: Mapped[str]

    real_remote_jid_: JID | None | ValueMissingT = dataclasses.field(
        default=VALUE_MISSING, repr=False
    )
    fk_real_remote_pk: Mapped[int | None] = mapped_column(
        StrValueMissingType, ForeignKey('remote.pk'), default=VALUE_MISSING, init=False
    )
    real_remote: Mapped[Remote | None] = relationship(
        lazy='joined',
        foreign_keys=fk_real_remote_pk,
        default=None,
        viewonly=True,
        init=False,
    )

    nickname: Mapped[str | None] = mapped_column(default=None)

    avatar_sha: Mapped[str | None] = mapped_column(
        StrValueMissingType, default=VALUE_MISSING
    )
    updated_at: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)

    def needs_update(self, existing: Occupant) -> bool:
        return existing.updated_at < self.updated_at

    def get_select_stmt(self) -> Select[Any]:
        return select(Occupant).where(
            Occupant.id == self.id,
            Occupant.fk_remote_pk == self.fk_remote_pk,
            Occupant.fk_account_pk == self.fk_account_pk,
        )

    def get_upsert_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {'updated_at': self.updated_at}
        if self.nickname is not None:
            values['nickname'] = self.nickname

        if isinstance(self.fk_real_remote_pk, int):
            values['fk_real_remote_pk'] = sa.func.coalesce(
                Occupant.fk_real_remote_pk, self.fk_real_remote_pk
            )

        if self.avatar_sha is not VALUE_MISSING:
            values['avatar_sha'] = self.avatar_sha

        return values


class OOB(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'oob'

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey('message.pk', ondelete='CASCADE'), init=False
    )
    url: Mapped[str]
    description: Mapped[str | None]


class Reply(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'reply'

    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey('message.pk', ondelete='CASCADE'), primary_key=True, init=False
    )
    id: Mapped[str]
    to: Mapped[JID | None] = mapped_column(JIDType)


class Call(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'call'

    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey('message.pk', ondelete='CASCADE'), primary_key=True, init=False
    )
    sid: Mapped[str]
    end_ts: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=None
    )
    state: Mapped[int]


class Encryption(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'encryption'
    __index_cols__ = ['key', 'trust', 'protocol']
    __no_table_cols__ = []
    __table_args__ = (Index('idx_encryption', 'key', 'trust', 'protocol', unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    protocol: Mapped[int]
    key: Mapped[str]
    trust: Mapped[int]

    def get_select_stmt(self) -> Select[Any]:
        return select(Encryption).where(
            Encryption.key == self.key,
            Encryption.trust == self.trust,
            Encryption.protocol == self.protocol,
        )


class SecurityLabel(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'securitylabel'
    __index_cols__ = ['label_hash', 'fk_remote_pk', 'fk_account_pk']
    __upsert_cols__ = ['displaymarking', 'fgcolor', 'bgcolor', 'updated_at']
    __no_table_cols__ = ['account_', 'remote_jid_']
    __table_args__ = (Index('idx_security_label', *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    label_hash: Mapped[str]
    displaymarking: Mapped[str]
    fgcolor: Mapped[str]
    bgcolor: Mapped[str]
    updated_at: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)

    def needs_update(self, existing: SecurityLabel) -> bool:
        return existing.updated_at < self.updated_at

    def get_select_stmt(self) -> Select[Any]:
        return select(SecurityLabel).where(
            SecurityLabel.label_hash == self.label_hash,
            SecurityLabel.fk_remote_pk == self.fk_remote_pk,
            SecurityLabel.fk_account_pk == self.fk_account_pk,
        )


class MessageError(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'error'
    __index_cols__ = ['message_id', 'fk_remote_pk', 'fk_account_pk']
    __no_table_cols__ = ['account_', 'remote_jid_']
    __table_args__ = (Index('idx_error', *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    message_id: Mapped[str]
    by: Mapped[JID | None] = mapped_column(JIDType)
    type: Mapped[str]
    text: Mapped[str | None]
    condition: Mapped[str]
    condition_text: Mapped[str | None]
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)


class Moderation(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'moderation'
    __index_cols__ = ['stanza_id', 'fk_remote_pk', 'fk_account_pk']
    __no_table_cols__ = ['account_', 'remote_jid_', 'occupant', 'occupant_']
    __table_args__ = (Index('idx_moderation', *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    occupant_: Occupant | None = dataclasses.field(repr=False)
    occupant: Mapped[Occupant | None] = relationship(
        lazy='joined', viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey('occupant.pk'), default=None, init=False
    )

    stanza_id: Mapped[str]
    by: Mapped[JID | None] = mapped_column(JIDType)
    reason: Mapped[str | None]
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)


class DisplayedMarker(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'displayed_marker'
    __no_table_cols__ = ['account_', 'remote_jid_', 'occupant', 'occupant_']

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    occupant_: Occupant | None = dataclasses.field(repr=False)
    occupant: Mapped[Occupant | None] = relationship(
        lazy='joined', viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey('occupant.pk'), default=None, init=False
    )

    id: Mapped[str] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)

    __table_args__ = (
        Index(
            'idx_displayed_marker',
            'id',
            'fk_remote_pk',
            'fk_account_pk',
            unique=True,
            sqlite_where=fk_occupant_pk.is_(None),
        ),
        Index(
            'idx_displayed_marker_gc',
            'id',
            'fk_remote_pk',
            'fk_occupant_pk',
            'fk_account_pk',
            unique=True,
            sqlite_where=fk_occupant_pk.isnot(None),
        ),
    )


class Receipt(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'receipt'
    __no_table_cols__ = ['account_', 'remote_jid_']
    __table_args__ = (
        Index(
            'idx_receipt',
            'id',
            'fk_remote_pk',
            'fk_account_pk',
            unique=True,
        ),
    )

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    id: Mapped[str] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)


class MAMArchiveState(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'mam_archive_state'
    __index_cols__ = ['fk_remote_pk', 'fk_account_pk']
    __upsert_cols__ = [
        'from_stanza_id',
        'from_stanza_ts',
        'to_stanza_id',
        'to_stanza_ts',
    ]
    __no_table_cols__ = ['account_', 'remote_jid_']
    __table_args__ = (Index('idx_mam_archive_state', *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    from_stanza_id: Mapped[str | None] = mapped_column(
        StrValueMissingType, default=VALUE_MISSING
    )
    from_stanza_ts: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=VALUE_MISSING
    )
    to_stanza_id: Mapped[str | None] = mapped_column(
        StrValueMissingType, default=VALUE_MISSING
    )
    to_stanza_ts: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=VALUE_MISSING
    )

    def get_select_stmt(self) -> Select[Any]:
        return select(MAMArchiveState).where(
            MAMArchiveState.fk_remote_pk == self.fk_remote_pk,
            MAMArchiveState.fk_account_pk == self.fk_account_pk,
        )

    def needs_update(self, _existing: MAMArchiveState) -> bool:
        return True


class FileTransferSource(MappedAsDataclass, Base, kw_only=True):
    __tablename__ = 'ft_source'

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)
    fk_filetransfer_pk: Mapped[int] = mapped_column(
        ForeignKey('filetransfer.pk', ondelete='CASCADE'), init=False
    )
    type: Mapped[str]

    __mapper_args__ = {
        'polymorphic_identity': 'source',
        'polymorphic_on': 'type',
    }


class UrlData(FileTransferSource):
    __tablename__ = 'ft_source_urldata'

    fk_ft_source_pk: Mapped[int] = mapped_column(
        ForeignKey('ft_source.pk', ondelete='CASCADE'), primary_key=True, init=False
    )
    target: Mapped[str]
    scheme_data: Mapped[dict[str, Any] | None] = mapped_column(JSONType)

    __mapper_args__ = {
        'polymorphic_load': 'selectin',
        'polymorphic_identity': 'urldata',
    }


class JingleFT(FileTransferSource):
    __tablename__ = 'ft_source_jingleft'

    fk_ft_source_pk: Mapped[int] = mapped_column(
        ForeignKey('ft_source.pk', ondelete='CASCADE'), primary_key=True, init=False
    )
    sid: Mapped[str]

    __mapper_args__ = {
        'polymorphic_load': 'selectin',
        'polymorphic_identity': 'jingleft',
    }


class JinglePub(FileTransferSource):
    __tablename__ = 'ft_source_jinglepub'

    fk_ft_source_pk: Mapped[int] = mapped_column(
        ForeignKey('ft_source.pk', ondelete='CASCADE'), primary_key=True, init=False
    )
    id: Mapped[str]

    __mapper_args__ = {
        'polymorphic_load': 'selectin',
        'polymorphic_identity': 'jinglepub',
    }


class FileTransfer(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'filetransfer'

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)
    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey('message.pk', ondelete='CASCADE'), init=False
    )

    date: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=None
    )
    desc: Mapped[str | None] = mapped_column(default=None)
    hash: Mapped[str | None] = mapped_column(default=None)
    hash_algo: Mapped[str | None] = mapped_column(default=None)
    height: Mapped[int | None] = mapped_column(default=None)
    length: Mapped[int | None] = mapped_column(default=None)
    media_type: Mapped[str | None] = mapped_column(default=None)
    name: Mapped[str | None] = mapped_column(default=None)
    size: Mapped[int | None] = mapped_column(default=None)
    width: Mapped[int | None] = mapped_column(default=None)

    state: Mapped[int | None]
    path: Mapped[str | None] = mapped_column(default=None, init=False)

    source: Mapped[list[FileTransferSource]] = relationship(
        lazy='selectin',
        default_factory=list,
        cascade='all, delete',
        passive_deletes=True,
    )


class Reaction(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'reaction'
    __upsert_cols__ = ['emojis', 'timestamp']
    __no_table_cols__ = ['account_', 'remote_jid_', 'occupant', 'occupant_']

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    occupant_: Occupant | None = dataclasses.field(repr=False)
    occupant: Mapped[Occupant | None] = relationship(
        lazy='joined', viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey('occupant.pk'), default=None, init=False
    )

    id: Mapped[str] = mapped_column()
    direction: Mapped[int] = mapped_column()
    emojis: Mapped[str] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)

    __table_args__ = (
        Index(
            'idx_reaction_id',
            'id',
            'fk_remote_pk',
            'fk_account_pk',
            'direction',
            unique=True,
            sqlite_where=fk_occupant_pk.is_(None),
        ),
        Index(
            'idx_reaction_id_gc',
            'id',
            'fk_remote_pk',
            'fk_occupant_pk',
            'fk_account_pk',
            'direction',
            unique=True,
            sqlite_where=fk_occupant_pk.isnot(None),
        ),
    )


class Message(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = 'message'

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    account: Mapped[Account] = relationship(lazy='joined', viewonly=True, init=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey('account.pk', ondelete='CASCADE'), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    remote: Mapped[Remote] = relationship(lazy='joined', viewonly=True, init=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey('remote.pk'), init=False)

    resource: Mapped[str | None] = mapped_column()
    type: Mapped[int] = mapped_column()
    direction: Mapped[int] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)
    state: Mapped[int]
    id: Mapped[str | None] = mapped_column(default=None)
    stanza_id: Mapped[str | None] = mapped_column(default=None)
    text: Mapped[str | None] = mapped_column(default=None)
    markup_type: Mapped[int | None] = mapped_column(default=None)
    markup: Mapped[str | None] = mapped_column(default=None)
    user_delay_ts: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=None
    )

    thread_id_: str | None = dataclasses.field(repr=False, default=None)
    thread: Mapped[Thread | None] = relationship(
        lazy='joined', viewonly=True, init=False
    )
    fk_thread_pk: Mapped[int | None] = mapped_column(
        ForeignKey('thread.pk'), default=None, init=False
    )

    occupant_: Occupant | None = dataclasses.field(repr=False, default=None)
    occupant: Mapped[Occupant | None] = relationship(
        lazy='joined', viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey('occupant.pk'), default=None, init=False
    )

    encryption_: Encryption | None = dataclasses.field(repr=False, default=None)
    encryption: Mapped[Encryption | None] = relationship(
        lazy='joined', viewonly=True, init=False
    )
    fk_encryption_pk: Mapped[int | None] = mapped_column(
        ForeignKey('encryption.pk'), default=None, init=False
    )

    security_label_: SecurityLabel | None = dataclasses.field(repr=False, default=None)
    security_label: Mapped[SecurityLabel | None] = relationship(
        lazy='joined', viewonly=True, init=False
    )
    fk_security_label_pk: Mapped[int | None] = mapped_column(
        ForeignKey('securitylabel.pk'), default=None, init=False
    )

    correction_id: Mapped[str | None] = mapped_column(default=None)

    corrections: Mapped[list[Message]] = relationship(
        lazy='selectin',
        init=False,
        primaryjoin=sa.and_(
            sa.orm.foreign(id) == sa.orm.remote(correction_id),
            sa.orm.foreign(fk_remote_pk) == sa.orm.remote(fk_remote_pk),
            sa.orm.foreign(fk_account_pk) == sa.orm.remote(fk_account_pk),
            sa.orm.foreign(fk_occupant_pk).is_(sa.orm.remote(fk_occupant_pk)),
            sa.orm.foreign(direction) == sa.orm.remote(direction),
            expr.case(
                (
                    sa.and_(
                        sa.orm.foreign(type) == 2,  # Groupchat
                        sa.orm.foreign(fk_occupant_pk).is_(None),
                    ),
                    sa.orm.foreign(resource) == sa.orm.remote(resource),
                ),
                else_=1,
            ),
        ),
        join_depth=1,
        viewonly=True,
        uselist=True,
        order_by=timestamp,
    )

    markers: Mapped[list[DisplayedMarker]] = relationship(
        lazy='selectin',
        init=False,
        primaryjoin=sa.and_(
            expr.case(
                (
                    type == 2,  # Groupchat
                    sa.orm.foreign(stanza_id) == DisplayedMarker.id,
                ),
                else_=sa.orm.foreign(id) == DisplayedMarker.id,
            ),
            fk_remote_pk == DisplayedMarker.fk_remote_pk,
            fk_account_pk == DisplayedMarker.fk_account_pk,
        ),
        viewonly=True,
        uselist=True,
    )

    receipt: Mapped[Receipt | None] = relationship(
        lazy='selectin',
        init=False,
        primaryjoin=sa.and_(
            sa.orm.foreign(id) == Receipt.id,
            fk_remote_pk == Receipt.fk_remote_pk,
            fk_account_pk == Receipt.fk_account_pk,
        ),
        viewonly=True,
    )

    reactions: Mapped[list[Reaction]] = relationship(
        lazy='selectin',
        init=False,
        primaryjoin=sa.and_(
            expr.case(
                (
                    type == 2,  # Groupchat
                    sa.orm.foreign(stanza_id) == Reaction.id,
                ),
                else_=sa.orm.foreign(id) == Reaction.id,
            ),
            fk_remote_pk == Reaction.fk_remote_pk,
            fk_account_pk == Reaction.fk_account_pk,
        ),
        viewonly=True,
        uselist=True,
    )

    moderation: Mapped[Moderation | None] = relationship(
        lazy='joined',
        default=None,
        init=False,
        primaryjoin=sa.and_(
            sa.orm.foreign(stanza_id) == sa.orm.remote(Moderation.stanza_id),
            fk_remote_pk == Moderation.fk_remote_pk,
            fk_account_pk == Moderation.fk_account_pk,
        ),
        viewonly=True,
    )

    error: Mapped[MessageError | None] = relationship(
        lazy='joined',
        default=None,
        init=False,
        primaryjoin=sa.and_(
            sa.orm.foreign(id) == sa.orm.remote(MessageError.message_id),
            fk_remote_pk == MessageError.fk_remote_pk,
            fk_account_pk == MessageError.fk_account_pk,
        ),
        viewonly=True,
    )

    filetransfers: Mapped[list[FileTransfer]] = relationship(
        lazy='selectin',
        default_factory=list,
        cascade='all, delete',
        passive_deletes=True,
    )
    call: Mapped[Call | None] = relationship(
        lazy='joined',
        default=None,
        cascade='all, delete',
        passive_deletes=True,
    )
    oob: Mapped[list[OOB]] = relationship(
        lazy='selectin',
        default_factory=list,
        cascade='all, delete',
        passive_deletes=True,
    )
    reply: Mapped[Reply | None] = relationship(
        lazy='joined',
        default=None,
        cascade='all, delete',
        passive_deletes=True,
    )

    __table_args__ = (
        Index(
            'idx_message', 'fk_remote_pk', 'fk_account_pk', sa.text('timestamp DESC')
        ),
        Index(
            'idx_message_corrections',
            'correction_id',
            'fk_remote_pk',
            'fk_account_pk',
            sa.text('timestamp ASC'),
        ),
        Index(
            'idx_message_id',
            'id',
            'fk_remote_pk',
            'fk_account_pk',
        ),
        Index(
            'idx_stanza_id',
            'stanza_id',
            'fk_remote_pk',
            'fk_account_pk',
        ),
    )

    def get_last_correction(self) -> Message:
        return self.corrections[-1]

    def get_referenced_message(self) -> Message | None:
        if self.reply is None:
            return None
        account = app.settings.get_account_from_jid(self.account.jid)
        return app.storage.archive.get_referenced_message(
            account,
            self.remote.jid,
            self.type,
            self.reply.id
        )



from typing import Any
from typing import Literal
from typing import overload

from gi.repository import Gtk
from gi.repository import GtkSource

class GajimBuilder:
    def __init__(
        self,
        filename: str | None = None,
        instance: Any = None,
        widgets: list[str] | None = None,
        domain: str | None = None,
        gettext_: Any | None = None,
    ) -> None: ...

class Builder(Gtk.Builder): ...

class AccountPageBuilder(Builder):
    scrolled: Gtk.ScrolledWindow
    account_box: Gtk.Box
    avatar_image: Gtk.Image
    account_label: Gtk.Label
    our_jid_label: Gtk.Label
    account_settings_button: Gtk.Button
    account_page_menu_button: Gtk.MenuButton
    status_box: Gtk.Box
    notifications_menu_button: Gtk.MenuButton

class AccountWizardBuilder(Builder):
    account_label_box: Gtk.Box
    account_name_entry: Gtk.Entry
    account_color_button: Gtk.ColorDialogButton
    badge_preview: Gtk.Label
    advanced_grid: Gtk.Grid
    custom_port_entry: Gtk.Entry
    custom_host_entry: Gtk.Entry
    proxies_combobox: Gtk.ComboBox
    manage_proxies_button: Gtk.Button
    con_type_combo: Gtk.ComboBoxText
    entrycompletion1: Gtk.EntryCompletion
    login_box: Gtk.Box
    log_in_address_entry: Gtk.Entry
    log_in_password_entry: Gtk.Entry
    login_advanced_checkbutton: Gtk.CheckButton
    log_in_button: Gtk.Button
    sign_up_button: Gtk.Button
    redirect_box: Gtk.Box
    instructions: Gtk.Label
    link_button: Gtk.Button
    security_warning_box: Gtk.Box
    error_list: Gtk.ListBox
    view_cert_button: Gtk.Button
    trust_cert_checkbutton: Gtk.CheckButton
    server_recommendations: Gtk.Popover
    visit_server_button: Gtk.Button
    recommendation_link1: Gtk.Label
    recommendation_link2: Gtk.Label
    signup_grid: Gtk.Grid
    sign_up_advanced_checkbutton: Gtk.CheckButton
    sign_up_anonymously: Gtk.CheckButton
    sign_up_info_grid: Gtk.Grid
    server_comboboxtext_sign_up: Gtk.ComboBoxText
    update_provider_list_icon: Gtk.Image

class AddContactBuilder(Builder):
    address_box: Gtk.Box
    account_box: Gtk.Box
    account_combo: Gtk.ComboBox
    cellrenderertext3: Gtk.CellRendererText
    address_entry: Gtk.Entry
    contact_grid: Gtk.Grid
    status_switch: Gtk.Switch
    group_combo: Gtk.ComboBoxText
    message_entry: Gtk.Entry
    contact_info_button: Gtk.Button
    gateway_box: Gtk.Box
    gateway_image: Gtk.Image
    gateway_label: Gtk.Label
    register_button: Gtk.Button
    commands_button: Gtk.Button

class AdvancedConfigurationBuilder(Builder):
    box: Gtk.Box
    search_entry: Gtk.SearchEntry
    advanced_treeview: Gtk.TreeView
    treeview_selection: Gtk.TreeSelection
    description: Gtk.Label
    reset_button: Gtk.Button

class AppPageBuilder(Builder):
    gajim_update: Gtk.Box
    update_message: Gtk.Label
    download_update: Gtk.Button
    dismiss_gajim_update: Gtk.Button
    gajim_update_check: Gtk.Box
    activate_update_check: Gtk.Button
    dismiss_update_check: Gtk.Button
    plugin_updates: Gtk.Box
    auto_update_plugins: Gtk.CheckButton
    update_plugins: Gtk.Button
    open_plugins: Gtk.Button
    dismiss_plugin_updates: Gtk.Button
    plugin_updates_finished: Gtk.Box
    notify_after_plugin_updates: Gtk.CheckButton
    dismiss_update_notification: Gtk.Button

class AssistantBuilder(Builder):
    main_grid: Gtk.Grid
    content_area: Gtk.Box
    stack: Gtk.Stack
    action_area: Gtk.Box

class BlockingListBuilder(Builder):
    blocking_grid: Gtk.Grid
    overlay: Gtk.Overlay
    block_view: Gtk.TreeView
    jid_cell_renderer: Gtk.CellRendererText
    add_button: Gtk.Button
    remove_button: Gtk.Button
    save_button: Gtk.Button

class CallWindowBuilder(Builder):
    adjustment1: Gtk.Adjustment
    adjustment2: Gtk.Adjustment
    dtmf_popover: Gtk.Popover
    grid1: Gtk.Grid
    button_1: Gtk.Button
    button_2: Gtk.Button
    button_3: Gtk.Button
    button_4: Gtk.Button
    button_5: Gtk.Button
    button_6: Gtk.Button
    button_7: Gtk.Button
    button_8: Gtk.Button
    button_9: Gtk.Button
    button_star: Gtk.Button
    button_0: Gtk.Button
    button_pound: Gtk.Button
    av_box: Gtk.Box
    video_box: Gtk.Box
    outgoing_viewport: Gtk.Viewport
    incoming_viewport: Gtk.Viewport
    avatar_image: Gtk.Image
    jingle_audio_state: Gtk.Image
    jingle_connection_state: Gtk.Label
    jingle_connection_spinner: Gtk.Spinner
    answer_video_button: Gtk.Button
    av_cam_button: Gtk.Button
    av_cam_image: Gtk.Image
    end_call_button: Gtk.Button
    audio_buttons_box: Gtk.Box
    mic_hscale: Gtk.VolumeButton
    volumebutton_plus_button2: Gtk.Button
    volumebutton_minus_button2: Gtk.Button
    sound_hscale: Gtk.VolumeButton
    volumebutton_plus_button1: Gtk.Button
    volumebutton_minus_button1: Gtk.Button
    dtmf_button: Gtk.MenuButton

class CertificateBuilder(Builder):
    certificate_box: Gtk.Box
    label_cert_for_account: Gtk.Label
    data_it_common_name: Gtk.Label
    data_it_organization: Gtk.Label
    data_it_subject_alt_names: Gtk.Label
    data_it_serial_number: Gtk.Label
    data_ib_common_name: Gtk.Label
    data_ib_organization: Gtk.Label
    data_issued_on: Gtk.Label
    data_expires_on: Gtk.Label
    data_sha1: Gtk.Label
    data_sha256: Gtk.Label
    copy_button: Gtk.Button
    image1: Gtk.Image
    public_key_algorithm: Gtk.Label
    public_key_size: Gtk.Label

class ChatBannerBuilder(Builder):
    share_popover: Gtk.Popover
    share_instructions: Gtk.Label
    qr_code_image: Gtk.Image
    jid_label: Gtk.Label
    copy_jid_button: Gtk.Button
    banner_box: Gtk.Box
    avatar_image: Gtk.Image
    name_label: Gtk.Label
    phone_image: Gtk.Image
    robot_image: Gtk.Image
    description_label: Gtk.Label
    additional_items_box: Gtk.Box
    share_menu_button: Gtk.MenuButton
    contact_info_button: Gtk.Button
    toggle_roster_button: Gtk.Button
    toggle_roster_image: Gtk.Image
    chat_menu_button: Gtk.MenuButton

class ChatControlBuilder(Builder):
    control_box: Gtk.Box
    conv_view_paned: Gtk.Paned
    conv_view_overlay: Gtk.Overlay

class ChatListRowBuilder(Builder):
    mainbox: Gtk.Box
    account_identifier: Gtk.Box
    avatar_image: Gtk.Image
    connection_icon: Gtk.Image
    group_chat_indicator: Gtk.Image
    name_label: Gtk.Label
    chatstate_image: Gtk.Image
    mute_image: Gtk.Image
    timestamp_label: Gtk.Label
    nick_label: Gtk.Label
    message_icon: Gtk.Image
    message_label: Gtk.Label
    unread_label: Gtk.Label
    revealer: Gtk.Revealer
    close_button: Gtk.Button

class ChatPanedBuilder(Builder):
    paned: Gtk.Paned
    middle_grid: Gtk.Grid
    header_bar: Gtk.Grid
    controls_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    start_chat_menu_button: Gtk.MenuButton
    section_label_eventbox: Gtk.Box
    section_label: Gtk.Label
    workspace_settings_button: Gtk.Button
    chat_list_scrolled: Gtk.ScrolledWindow
    right_grid: Gtk.Grid

class ContactInfoBuilder(Builder):
    devices_grid: Gtk.Grid
    priority_label: Gtk.Label
    software_label: Gtk.Label
    system_label: Gtk.Label
    time_label: Gtk.Label
    status_value: Gtk.Label
    priority_value: Gtk.Label
    software_value: Gtk.Label
    system_value: Gtk.Label
    time_value: Gtk.Label
    resource_box: Gtk.Box
    resource_label: Gtk.Label
    groups_model: Gtk.ListStore
    main_grid: Gtk.Grid
    header_revealer: Gtk.Revealer
    avatar_image_header: Gtk.Image
    contact_name_header_label: Gtk.Label
    edit_contact_name_header_button: Gtk.Button
    main_stack: Gtk.Stack
    vcard_scrolled: Gtk.ScrolledWindow
    vcard_box: Gtk.Box
    avatar_image: Gtk.Image
    contact_name_controls_box: Gtk.Box
    contact_jid_label: Gtk.Label
    group_chat_grid: Gtk.Grid
    role_label: Gtk.Label
    affiliation_label: Gtk.Label
    settings_scrolled: Gtk.ScrolledWindow
    settings_box: Gtk.Box
    subscription_listbox: Gtk.ListBox
    from_subscription_switch: Gtk.Switch
    to_subscription_stack: Gtk.Stack
    request_stack: Gtk.Stack
    to_subscription_button: Gtk.Button
    contact_settings_box: Gtk.Box
    remove_history_button: Gtk.Button
    encryption_scrolled: Gtk.ScrolledWindow
    encryption_box: Gtk.Box
    groups_page_stack: Gtk.Stack
    groups_treeview: Gtk.TreeView
    tree_selection: Gtk.TreeSelection
    toggle_renderer: Gtk.CellRendererToggle
    text_renderer: Gtk.CellRendererText
    group_add_button: Gtk.Button
    group_remove_button: Gtk.Button
    notes_page_stack: Gtk.Stack
    scrolledwindow_annotation: Gtk.ScrolledWindow
    textview_annotation: Gtk.TextView
    devices_stack: Gtk.Stack
    devices_box: Gtk.Box

class ContactTooltipBuilder(Builder):
    tooltip_grid: Gtk.Grid
    jid: Gtk.Label
    tune_label: Gtk.Label
    location_label: Gtk.Label
    tune: Gtk.Label
    location: Gtk.Label
    name: Gtk.Label
    avatar: Gtk.Image
    sub_label: Gtk.Label
    sub: Gtk.Label
    resources_box: Gtk.Box

class DbMigrationBuilder(Builder):
    box: Gtk.Box
    stack: Gtk.Stack
    status_label: Gtk.Label
    error_label: Gtk.Label
    error_view: Gtk.TextView
    error_copy_button: Gtk.Button
    error_close_button: Gtk.Button
    success_close_button: Gtk.Button

class DebugConsoleBuilder(Builder):
    popover: Gtk.Popover
    stanza_presets_listbox: Gtk.ListBox
    stack: Gtk.Stack
    log_view: GtkSource.View
    paned: Gtk.Paned
    search_revealer: Gtk.Revealer
    search_entry: Gtk.SearchEntry
    search_forward: Gtk.Button
    search_backward: Gtk.Button
    search_results_label: Gtk.Label
    scrolled: Gtk.ScrolledWindow
    protocol_view: GtkSource.View
    jump_to_end_button: Gtk.Button
    scrolled_input: Gtk.ScrolledWindow
    input_entry: GtkSource.View
    actionbox: Gtk.Box
    filter_options_button: Gtk.Button
    clear_button: Gtk.Button
    edit_toggle: Gtk.ToggleButton
    paste: Gtk.Button
    menubutton: Gtk.MenuButton
    account_label: Gtk.Label
    send: Gtk.Button
    headerbar: Gtk.HeaderBar
    search_toggle: Gtk.ToggleButton

class ExceptionDialogBuilder(Builder):
    exception_box: Gtk.Box
    infobar: Gtk.Revealer
    exception_view: Gtk.TextView
    user_feedback_box: Gtk.Box
    user_feedback_entry: Gtk.Entry
    close_button: Gtk.Button
    report_spinner: Gtk.Spinner
    report_button: Gtk.Button

class FileTransferBuilder(Builder):
    transfer_box: Gtk.Box
    transfer_description: Gtk.Label
    file_name: Gtk.Label
    file_size: Gtk.Label
    progress_bar: Gtk.ProgressBar
    transfer_progress: Gtk.Label
    cancel_button: Gtk.Button

class FileTransferJingleBuilder(Builder):
    transfer_box: Gtk.Box
    transfer_action: Gtk.Label
    file_name: Gtk.Label
    file_description: Gtk.Label
    file_size: Gtk.Label
    action_stack: Gtk.Stack
    accept_file_request: Gtk.Button
    reject_file_request: Gtk.Button
    open_folder: Gtk.Button
    open_file: Gtk.Button
    error_show_transfers: Gtk.Button
    error_label: Gtk.Label
    retry_bad_hash: Gtk.Button
    rejected_show_transfers: Gtk.Button
    progress_label: Gtk.Label
    progress_bar: Gtk.ProgressBar
    cancel_transfer: Gtk.Button

class FileTransferSelectorBuilder(Builder):
    file_box: Gtk.Box
    preview_image_box: Gtk.Box
    preview_image: Gtk.Image
    file_name_label: Gtk.Label
    file_size_label: Gtk.Label
    warning_label: Gtk.Label
    remove_file_button: Gtk.Button
    stack: Gtk.Stack
    box: Gtk.Box
    listbox: Gtk.ListBox
    resource_box: Gtk.Box
    resource_instructions: Gtk.Label

class GroupchatAffiliationBuilder(Builder):
    affiliation_store: Gtk.ListStore
    combo_store: Gtk.ListStore
    main_box: Gtk.Box
    affiliation_scrolled: Gtk.ScrolledWindow
    affiliation_treeview: Gtk.TreeView
    affiliation_selection: Gtk.TreeSelection
    address_renderer: Gtk.CellRendererText
    reserved_name_column: Gtk.TreeViewColumn
    reserved_name_renderer: Gtk.CellRendererText
    affiliation_renderer: Gtk.CellRendererCombo
    button_box: Gtk.Box
    add_remove_button_box: Gtk.Box
    add_button: Gtk.Button
    remove_button: Gtk.Button

class GroupchatConfigBuilder(Builder):
    stack: Gtk.Stack
    loading_box: Gtk.Box
    config_box: Gtk.Box
    error_box: Gtk.Box
    error_image: Gtk.Image
    error_label: Gtk.Label

class GroupchatCreationBuilder(Builder):
    account_liststore: Gtk.ListStore
    stack: Gtk.Stack
    grid: Gtk.Grid
    name_entry: Gtk.Entry
    description_entry: Gtk.Entry
    account_combo: Gtk.ComboBox
    account_label: Gtk.Label
    advanced_switch: Gtk.Switch
    advanced_switch_label: Gtk.Label
    error_label: Gtk.Label
    info_label: Gtk.Label
    address_entry_label: Gtk.Label
    address_entry: Gtk.Entry
    public_radio: Gtk.CheckButton
    private_radio: Gtk.CheckButton
    spinner: Gtk.Spinner
    create_button: Gtk.Button

class GroupchatDetailsBuilder(Builder):
    main_grid: Gtk.Grid
    header_revealer: Gtk.Revealer
    header_image: Gtk.Image
    contact_name_header_label: Gtk.Label
    edit_name_button: Gtk.Button
    main_stack: Gtk.Stack
    info_box: Gtk.Box
    settings_box: Gtk.Box
    encryption_scrolled: Gtk.ScrolledWindow
    encryption_box: Gtk.Box
    manage_box: Gtk.Box
    affiliation_box: Gtk.Box
    outcasts_box: Gtk.Box
    configuration_box: Gtk.Box

class GroupchatInfoScrolledBuilder(Builder):
    info_grid: Gtk.Grid
    address_label: Gtk.Label
    description_label: Gtk.Label
    subject_label: Gtk.Label
    author_label: Gtk.Label
    description: Gtk.Label
    author: Gtk.Label
    users: Gtk.Label
    contact_label: Gtk.Label
    logs_label: Gtk.Label
    lang: Gtk.Label
    logs: Gtk.LinkButton
    users_image: Gtk.Image
    lang_image: Gtk.Image
    contact_box: Gtk.Box
    subject: Gtk.Label
    avatar_image: Gtk.Image
    address: Gtk.Label
    address_copy_button: Gtk.Button
    name_box: Gtk.Box

class GroupchatInviterBuilder(Builder):
    account_store: Gtk.ListStore
    invite_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    scrolledwindow: Gtk.ScrolledWindow
    contacts_listbox: Gtk.ListBox
    contacts_placeholder: Gtk.Box
    invitees_scrolled: Gtk.ScrolledWindow

class GroupchatManageBuilder(Builder):
    stack: Gtk.Stack
    avatar_overlay: Gtk.Overlay
    avatar_button_image: Gtk.Image
    remove_avatar_button: Gtk.Button
    muc_description_entry: Gtk.Entry
    muc_name_entry: Gtk.Entry
    destroy_muc_button: Gtk.Button
    manage_save_button: Gtk.Button
    subject_textview: Gtk.TextView
    subject_change_button: Gtk.Button
    avatar_selector_grid: Gtk.Grid
    avatar_cancel_button: Gtk.Button
    avatar_update_button: Gtk.Button
    destroy_reason_entry: Gtk.Entry
    destroy_alternate_entry: Gtk.Entry
    destroy_cancel_button: Gtk.Button
    destroy_button: Gtk.Button

class GroupchatNickChooserBuilder(Builder):
    button_content: Gtk.Box
    label: Gtk.Label
    popover: Gtk.Popover
    entry: Gtk.Entry
    apply_button: Gtk.Button

class GroupchatOutcastBuilder(Builder):
    info_popover: Gtk.Popover
    outcast_store: Gtk.ListStore
    main_box: Gtk.Box
    outcast_scrolled: Gtk.ScrolledWindow
    outcast_treeview: Gtk.TreeView
    outcast_selection: Gtk.TreeSelection
    address_renderer: Gtk.CellRendererText
    reason_renderer: Gtk.CellRendererText
    button_box: Gtk.Box
    add_remove_button_box: Gtk.Box
    add_button: Gtk.Button
    remove_button: Gtk.Button
    info_button: Gtk.MenuButton

class GroupchatRosterBuilder(Builder):
    box: Gtk.Box
    search_entry: Gtk.SearchEntry
    scrolled: Gtk.ScrolledWindow

class GroupchatRosterTooltipBuilder(Builder):
    tooltip_grid: Gtk.Grid
    avatar: Gtk.Image
    jid: Gtk.Label
    nick: Gtk.Label
    fillelement: Gtk.Label
    status: Gtk.Label
    affiliation: Gtk.Label
    hats_box: Gtk.Box

class GroupchatStateBuilder(Builder):
    groupchat_state: Gtk.Stack
    join_button: Gtk.Button
    joining_spinner: Gtk.Spinner
    abort_join_button: Gtk.Button
    mam_sync_spinner: Gtk.Spinner
    mam_error_label: Gtk.Label
    close_button: Gtk.Button

class HistoryExportBuilder(Builder):
    select_account_box: Gtk.Box
    settings_grid: Gtk.Grid
    account_combo: Gtk.ComboBox

class MainBuilder(Builder):
    main_grid: Gtk.Grid
    left_grid: Gtk.Grid
    workspace_scrolled: Gtk.ScrolledWindow
    app_box: Gtk.Box
    toggle_chat_list_button: Gtk.Button
    toggle_chat_list_icon: Gtk.Image
    account_box: Gtk.Box

class MamPreferencesBuilder(Builder):
    mam_box: Gtk.Box
    default_combo: Gtk.ComboBox
    overlay: Gtk.Overlay
    pref_view: Gtk.TreeView
    jid_cell_renderer: Gtk.CellRendererText
    pref_toggle_cell_renderer: Gtk.CellRendererToggle
    add: Gtk.Button
    remove: Gtk.Button
    save_button: Gtk.Button

class ManageProxiesBuilder(Builder):
    box: Gtk.Box
    proxies_treeview: Gtk.TreeView
    treeview_selection1: Gtk.TreeSelection
    add_proxy_button: Gtk.Button
    remove_proxy_button: Gtk.Button
    settings_grid: Gtk.Grid
    proxypass_entry: Gtk.Entry
    proxyuser_entry: Gtk.Entry
    useauth_checkbutton: Gtk.CheckButton
    proxyport_entry: Gtk.Entry
    proxyhost_entry: Gtk.Entry
    proxytype_combobox: Gtk.ComboBox
    proxyname_entry: Gtk.Entry

class ManageRosterBuilder(Builder):
    main: Gtk.Box
    top_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    scrolled_box: Gtk.Box
    scrolled: Gtk.ScrolledWindow
    column_view: Gtk.ColumnView
    jid_col: Gtk.ColumnViewColumn
    name_col: Gtk.ColumnViewColumn
    subscription_col: Gtk.ColumnViewColumn
    ask_col: Gtk.ColumnViewColumn
    import_button: Gtk.MenuButton
    export_button: Gtk.Button

class ManageSoundsBuilder(Builder):
    manage_sounds: Gtk.Box
    sounds_treeview: Gtk.TreeView
    toggle_cell_renderer: Gtk.CellRendererToggle
    sound_buttons_box: Gtk.Box
    clear_sound_button: Gtk.Button
    play_sound_button: Gtk.Button

class MessageActionsBoxBuilder(Builder):
    box: Gtk.Box
    chat_state_box: Gtk.Box
    reply_box: Gtk.Box
    state_box: Gtk.Box
    state_box_image: Gtk.Image
    state_box_label: Gtk.Label
    visitor_menu_button: Gtk.MenuButton
    edit_box: Gtk.Box
    edit_box_image: Gtk.Image
    cancel_correction_button: Gtk.Button
    action_box: Gtk.Box
    emoticons_button: Gtk.MenuButton
    formattings_button: Gtk.MenuButton
    input_wrapper: Gtk.Overlay
    input_scrolled: Gtk.ScrolledWindow
    input_overlay: Gtk.Box
    input_overlay_label: Gtk.Label
    send_message_button: Gtk.Button
    sendfile_button: Gtk.Button
    encryption_menu_button: Gtk.MenuButton
    encryption_image: Gtk.Image
    encryption_details_button: Gtk.Button
    encryption_details_image: Gtk.Image
    visitor_popover: Gtk.Popover
    request_voice_button: Gtk.Button

class OmemoTrustManagerBuilder(Builder):
    search_popover: Gtk.Popover
    search: Gtk.SearchEntry
    qr_code_popover: Gtk.Popover
    comparing_instructions: Gtk.Label
    our_fingerprint_2: Gtk.Label
    qr_code_image: Gtk.Image
    stack: Gtk.Stack
    our_fingerprint_1: Gtk.Label
    qr_menu_button: Gtk.MenuButton
    manage_trust_button: Gtk.Button
    list_heading: Gtk.Label
    list_heading_box: Gtk.Box
    show_inactive_switch: Gtk.Switch
    search_button: Gtk.MenuButton
    list: Gtk.ListBox
    clear_devices_button: Gtk.Button

class PasswordDialogBuilder(Builder):
    pass_box: Gtk.Box
    header: Gtk.Label
    message_label: Gtk.Label
    pass_entry: Gtk.PasswordEntry
    save_pass_checkbutton: Gtk.CheckButton
    keyring_hint: Gtk.Label
    cancel_button: Gtk.Button
    ok_button: Gtk.Button

class PepConfigBuilder(Builder):
    stack: Gtk.Stack
    overview_box: Gtk.Box
    services_treeview: Gtk.TreeView
    delete_button: Gtk.Button
    configure_button: Gtk.Button
    show_content_button: Gtk.Button
    items_box: Gtk.Box
    items_label: Gtk.Label
    items_view: GtkSource.View
    items_back_button: Gtk.Button
    config_box: Gtk.Box
    form_label: Gtk.Label
    form_box: Gtk.Box
    config_back_button: Gtk.Button
    save_button: Gtk.Button

class PluginsBuilder(Builder):
    liststore: Gtk.ListStore
    plugins_box: Gtk.Box
    plugins_treeview: Gtk.TreeView
    treeview_selection: Gtk.TreeSelection
    enabled_column: Gtk.TreeViewColumn
    enabled_renderer: Gtk.CellRendererToggle
    toolbar: Gtk.Box
    download_button: Gtk.Button
    uninstall_plugin_button: Gtk.Button
    help_button: Gtk.Button
    plugin_name_label: Gtk.Label
    configure_plugin_button: Gtk.Button
    description: Gtk.Label
    plugin_version_label: Gtk.Label
    plugin_authors_label: Gtk.Label
    plugin_homepage_linkbutton: Gtk.Label

class PreferencesBuilder(Builder):
    grid: Gtk.Grid
    stack: Gtk.Stack
    window_behaviour: Gtk.Grid
    plugins: Gtk.Grid
    general: Gtk.Grid
    chats: Gtk.Grid
    group_chats: Gtk.Grid
    file_preview: Gtk.Grid
    visual_notifications: Gtk.Grid
    sounds: Gtk.Grid
    status_message: Gtk.Grid
    automatic_status: Gtk.Grid
    themes: Gtk.Grid
    av_info_bar: Gtk.Revealer
    av_info_bar_label: Gtk.Label
    server: Gtk.Grid
    audio: Gtk.Grid
    video: Gtk.Grid
    miscellaneous: Gtk.Grid
    reset_button: Gtk.Button
    purge_history_button: Gtk.Button
    advanced: Gtk.Grid
    ace_button: Gtk.Button

class PreviewBuilder(Builder):
    preview_stack: Gtk.Stack
    preview_box: Gtk.Box
    icon_event_box: Gtk.Box
    icon_button: Gtk.Button
    right_box: Gtk.Box
    progress_box: Gtk.Box
    progressbar: Gtk.ProgressBar
    progress_text: Gtk.Label
    cancel_download_button: Gtk.Button
    content_box: Gtk.Box
    content_overlay: Gtk.Overlay
    image_button: Gtk.Button
    link_button: Gtk.LinkButton
    button_box: Gtk.Box
    file_name: Gtk.Label
    file_size: Gtk.Label
    info_message: Gtk.Label
    open_folder_button: Gtk.Button
    save_as_button: Gtk.Button
    download_button: Gtk.Button

class PreviewAudioBuilder(Builder):
    seek_bar_adj: Gtk.Adjustment
    speed_bar_adj: Gtk.Adjustment
    preview_box: Gtk.Box
    drawing_box: Gtk.Box
    seek_bar_box: Gtk.Box
    seek_bar: Gtk.Scale
    progress_label: Gtk.Label
    control_box: Gtk.Box
    rewind_button: Gtk.Button
    play_pause_button: Gtk.Button
    play_icon: Gtk.Image
    forward_button: Gtk.Button
    speed_dec_button: Gtk.Button
    speed_menubutton: Gtk.MenuButton
    speed_label: Gtk.Label
    speed_inc_button: Gtk.Button
    speed_popover: Gtk.Popover
    speed_bar: Gtk.Scale

class ProfileBuilder(Builder):
    privacy_popover: Gtk.Popover
    avatar_nick_access: Gtk.Switch
    vcard_access: Gtk.Switch
    avatar_nick_access_label: Gtk.Label
    vcard_access_label: Gtk.Label
    profile_stack: Gtk.Stack
    spinner: Gtk.Spinner
    scrolled: Gtk.ScrolledWindow
    profile_box: Gtk.Box
    avatar_overlay: Gtk.Overlay
    avatar_image: Gtk.Image
    remove_avatar_button: Gtk.Button
    nickname_entry: Gtk.Entry
    cancel_button: Gtk.Button
    add_entry_button: Gtk.MenuButton
    privacy_button: Gtk.MenuButton
    save_button: Gtk.Button
    edit_button: Gtk.Button
    avatar_selector_box: Gtk.Box
    avatar_cancel: Gtk.Button
    avatar_update_button: Gtk.Button
    error_label: Gtk.Label
    error_title_label: Gtk.Label
    back_button: Gtk.Button

class QuitDialogBuilder(Builder):
    box: Gtk.Box
    remember_checkbutton: Gtk.CheckButton
    hide_button: Gtk.Button
    minimize_button: Gtk.Button
    quit_button: Gtk.Button

class RosterItemExchangeBuilder(Builder):
    roster_item_exchange: Gtk.Box
    type_label: Gtk.Label
    body_scrolledwindow: Gtk.ScrolledWindow
    body_textview: Gtk.TextView
    items_list_treeview: Gtk.TreeView
    treeview_selection1: Gtk.TreeSelection
    cancel_button: Gtk.Button
    accept_button: Gtk.Button

class SearchViewBuilder(Builder):
    calendar_popover: Gtk.Popover
    calendar: Gtk.Calendar
    first_day_button: Gtk.Button
    previous_day_button: Gtk.Button
    next_day_button: Gtk.Button
    last_day_button: Gtk.Button
    search_box: Gtk.Box
    close_button: Gtk.Button
    calendar_button: Gtk.MenuButton
    search_entry: Gtk.SearchEntry
    search_checkbutton: Gtk.CheckButton
    search_filters_box: Gtk.Box
    results_scrolled: Gtk.ScrolledWindow
    results_listbox: Gtk.ListBox
    placeholder: Gtk.Box
    placeholder_image: Gtk.Image
    placeholder_label: Gtk.Label
    filter_date_before_popover: Gtk.Popover
    filter_date_before_calendar: Gtk.Calendar
    filter_date_before_reset_button: Gtk.Button
    filter_date_after_popover: Gtk.Popover
    filter_date_after_calendar: Gtk.Calendar
    filter_date_after_reset_button: Gtk.Button
    search_filters_grid: Gtk.Grid
    filter_from_desc_label: Gtk.Label
    filter_from_entry: Gtk.Entry
    filter_before_button: Gtk.MenuButton
    filter_after_button: Gtk.MenuButton
    filter_before_label: Gtk.Label
    filter_after_label: Gtk.Label
    header_box: Gtk.Box
    header_name_label: Gtk.Label
    header_date_label: Gtk.Label
    result_row_grid: Gtk.Grid
    row_avatar: Gtk.Image
    row_time_label: Gtk.Label
    row_name_label: Gtk.Label

class ServerInfoBuilder(Builder):
    server_info_notebook: Gtk.Notebook
    server: Gtk.Grid
    server_hostname: Gtk.Label
    server_software: Gtk.Label
    server_uptime: Gtk.Label
    no_addresses_label: Gtk.Label
    connection_type: Gtk.Label
    proxy_type: Gtk.Label
    proxy_host: Gtk.Label
    domain_label: Gtk.Label
    dns_label: Gtk.Label
    ip_port_label: Gtk.Label
    websocket_label: Gtk.Label
    domain: Gtk.Label
    dns: Gtk.Label
    ip_port: Gtk.Label
    websocket: Gtk.Label
    tls_version: Gtk.Label
    cipher_suite: Gtk.Label
    cert_scrolled: Gtk.ScrolledWindow
    no_certificate_label: Gtk.Label
    features: Gtk.Box
    features_listbox: Gtk.ListBox
    clipboard_button: Gtk.Button

class ServiceDiscoveryWindowBuilder(Builder):
    service_discovery: Gtk.Box
    banner_agent_icon: Gtk.Image
    banner_agent_header: Gtk.Label
    banner_agent_subheader: Gtk.Label
    address_box: Gtk.Box
    address_comboboxtext: Gtk.ComboBoxText
    browse_button: Gtk.Button
    services_progressbar: Gtk.ProgressBar
    services_scrollwin: Gtk.ScrolledWindow
    services_treeview: Gtk.TreeView
    treeview_selection1: Gtk.TreeSelection
    action_buttonbox: Gtk.Box

class ShortcutsWindowBuilder(Builder):
    shortcuts_window: Gtk.ShortcutsWindow

class SslErrorDialogBuilder(Builder):
    ssl_error_box: Gtk.Box
    intro_text: Gtk.Label
    ssl_error: Gtk.Label
    add_certificate_checkbutton: Gtk.CheckButton
    view_cert_button: Gtk.Button
    connect_button: Gtk.Button

class StartChatDialogBuilder(Builder):
    stack: Gtk.Stack
    infobar: Gtk.Revealer
    infobar_close_button: Gtk.Button
    box: Gtk.Box
    controls_box: Gtk.Box
    search_entry: Gtk.SearchEntry
    global_search_toggle: Gtk.ToggleButton
    search_error_box: Gtk.Box
    search_error_label: Gtk.Label
    list_stack: Gtk.Stack
    contact_scrolled: Gtk.ScrolledWindow
    no_contacts_placeholder: Gtk.Box
    global_scrolled: Gtk.ScrolledWindow
    global_search_placeholder_stack: Gtk.Stack
    global_search_placeholder_hints: Gtk.Box
    global_search_results_label: Gtk.Label
    spinner: Gtk.Spinner
    error_label: Gtk.Label
    error_back_button: Gtk.Button
    info_box: Gtk.Box
    info_back_button: Gtk.Button
    join_box: Gtk.Box
    join_button: Gtk.Button
    account_view: Gtk.TreeView
    icon_pixbuf: Gtk.CellRendererPixbuf
    account_text: Gtk.CellRendererText
    account_back_button: Gtk.Button
    account_select_button: Gtk.Button

class ThemesWindowBuilder(Builder):
    option_popover: Gtk.Popover
    choose_option_listbox: Gtk.ListBox
    placeholder: Gtk.Box
    theme_store: Gtk.ListStore
    theme_grid: Gtk.Grid
    theme_treeview: Gtk.TreeView
    theme_treeview_selection: Gtk.TreeSelection
    theme_name_cell_renderer: Gtk.CellRendererText
    option_listbox: Gtk.ListBox
    add_option_button: Gtk.MenuButton
    add_theme_button: Gtk.Button
    remove_theme_button: Gtk.Button

class VideoPreviewBuilder(Builder):
    video_preview_box: Gtk.Box
    video_source_label: Gtk.Label
    video_preview_placeholder: Gtk.Box

class VoiceMessageRecorderBuilder(Builder):
    popover: Gtk.Popover
    box: Gtk.Box
    error_label: Gtk.Label
    progression_box: Gtk.Box
    time_label: Gtk.Label
    visualization_box: Gtk.Box
    audio_player_box: Gtk.Box
    record_control_box: Gtk.Box
    cancel_button: Gtk.Button
    record_toggle_button: Gtk.Button
    record_toggle_button_image: Gtk.Image
    send_button: Gtk.Button

class WorkspaceDialogBuilder(Builder):
    box: Gtk.Box
    preview: Gtk.Image
    entry: Gtk.Entry
    remove_workspace_button: Gtk.Button
    image_switch: Gtk.Switch
    style_stack: Gtk.Stack
    color_dialog_button: Gtk.ColorDialogButton
    image_box: Gtk.Box
    cancel_button: Gtk.Button
    save_button: Gtk.Button

@overload
def get_builder(
    file_name: Literal["account_page.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> AccountPageBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["account_wizard.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> AccountWizardBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["add_contact.ui"], instance: Any = None, widgets: list[str] = ...
) -> AddContactBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["advanced_configuration.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> AdvancedConfigurationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["app_page.ui"], instance: Any = None, widgets: list[str] = ...
) -> AppPageBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["assistant.ui"], instance: Any = None, widgets: list[str] = ...
) -> AssistantBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["blocking_list.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> BlockingListBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["call_window.ui"], instance: Any = None, widgets: list[str] = ...
) -> CallWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["certificate.ui"], instance: Any = None, widgets: list[str] = ...
) -> CertificateBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["chat_banner.ui"], instance: Any = None, widgets: list[str] = ...
) -> ChatBannerBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["chat_control.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ChatControlBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["chat_list_row.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ChatListRowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["chat_paned.ui"], instance: Any = None, widgets: list[str] = ...
) -> ChatPanedBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["contact_info.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ContactInfoBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["contact_tooltip.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ContactTooltipBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["db_migration.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> DbMigrationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["debug_console.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> DebugConsoleBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["exception_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ExceptionDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer_jingle.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferJingleBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["file_transfer_selector.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> FileTransferSelectorBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_affiliation.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatAffiliationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_config.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatConfigBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_creation.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatCreationBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_details.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatDetailsBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_info_scrolled.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatInfoScrolledBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_inviter.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatInviterBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_manage.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatManageBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_nick_chooser.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatNickChooserBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_outcast.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatOutcastBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_roster.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatRosterBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_roster_tooltip.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatRosterTooltipBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["groupchat_state.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> GroupchatStateBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["history_export.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> HistoryExportBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["main.ui"], instance: Any = None, widgets: list[str] = ...
) -> MainBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["mam_preferences.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> MamPreferencesBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["manage_proxies.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ManageProxiesBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["manage_roster.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ManageRosterBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["manage_sounds.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ManageSoundsBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["message_actions_box.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> MessageActionsBoxBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["omemo_trust_manager.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> OmemoTrustManagerBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["password_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> PasswordDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["pep_config.ui"], instance: Any = None, widgets: list[str] = ...
) -> PepConfigBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["plugins.ui"], instance: Any = None, widgets: list[str] = ...
) -> PluginsBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["preferences.ui"], instance: Any = None, widgets: list[str] = ...
) -> PreferencesBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["preview.ui"], instance: Any = None, widgets: list[str] = ...
) -> PreviewBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["preview_audio.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> PreviewAudioBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["profile.ui"], instance: Any = None, widgets: list[str] = ...
) -> ProfileBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["quit_dialog.ui"], instance: Any = None, widgets: list[str] = ...
) -> QuitDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["roster_item_exchange.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> RosterItemExchangeBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["search_view.ui"], instance: Any = None, widgets: list[str] = ...
) -> SearchViewBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["server_info.ui"], instance: Any = None, widgets: list[str] = ...
) -> ServerInfoBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["service_discovery_window.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ServiceDiscoveryWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["shortcuts_window.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ShortcutsWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["ssl_error_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> SslErrorDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["start_chat_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> StartChatDialogBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["themes_window.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> ThemesWindowBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["video_preview.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> VideoPreviewBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["voice_message_recorder.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> VoiceMessageRecorderBuilder: ...  # noqa
@overload
def get_builder(
    file_name: Literal["workspace_dialog.ui"],
    instance: Any = None,
    widgets: list[str] = ...,
) -> WorkspaceDialogBuilder: ...  # noqa
def get_builder(
    file_name: str, instance: Any = None, widgets: list[str] = ...
) -> Builder: ...



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import sys

import gajim


def get_extended_app_version() -> str:
    if gajim.IS_FLATPAK:
        package = 'Flatpak'
    if gajim.IS_FLATPAK_NIGHTLY:
        package = 'Flatpak Nightly'
    elif gajim.IS_PORTABLE:
        package = 'Windows Portable'
    elif gajim.IS_MS_STORE:
        package = 'Windows Store'
    elif sys.platform == 'win32':
        package = 'Windows'
    elif sys.platform == 'darwin':
        package = 'MacOS'
    else:
        package = 'Unix/Linux'

    return f'{gajim.__version__} ({package})'



# Generated by get_iana_data.py @ 2025-02-27T18:57:21.073060+00:00

from typing import NamedTuple


class ZoneData(NamedTuple):
    key: str
    full_name: str
    short_name: str
    country_code: str
    country_name: str


def get_zone_data(key: str) -> ZoneData:
    return ZoneData(*ZONE_DATA[key])


ZONE_DATA = {
    'Africa/Abidjan': ('Africa/Abidjan', 'Africa/Abidjan', 'Abidjan', 'CI', "Côte d'Ivoire"),
    'Africa/Accra': ('Africa/Accra', 'Africa/Accra', 'Accra', 'GH', 'Ghana'),
    'Africa/Addis_Ababa': ('Africa/Addis_Ababa', 'Africa/Addis Ababa', 'Addis Ababa', 'ET', 'Ethiopia'),
    'Africa/Algiers': ('Africa/Algiers', 'Africa/Algiers', 'Algiers', 'DZ', 'Algeria'),
    'Africa/Asmara': ('Africa/Asmara', 'Africa/Asmara', 'Asmara', 'ER', 'Eritrea'),
    'Africa/Bamako': ('Africa/Bamako', 'Africa/Bamako', 'Bamako', 'ML', 'Mali'),
    'Africa/Bangui': ('Africa/Bangui', 'Africa/Bangui', 'Bangui', 'CF', 'Central African Rep.'),
    'Africa/Banjul': ('Africa/Banjul', 'Africa/Banjul', 'Banjul', 'GM', 'Gambia'),
    'Africa/Bissau': ('Africa/Bissau', 'Africa/Bissau', 'Bissau', 'GW', 'Guinea-Bissau'),
    'Africa/Blantyre': ('Africa/Blantyre', 'Africa/Blantyre', 'Blantyre', 'MW', 'Malawi'),
    'Africa/Brazzaville': ('Africa/Brazzaville', 'Africa/Brazzaville', 'Brazzaville', 'CG', 'Congo (Rep.)'),
    'Africa/Bujumbura': ('Africa/Bujumbura', 'Africa/Bujumbura', 'Bujumbura', 'BI', 'Burundi'),
    'Africa/Cairo': ('Africa/Cairo', 'Africa/Cairo', 'Cairo', 'EG', 'Egypt'),
    'Africa/Casablanca': ('Africa/Casablanca', 'Africa/Casablanca', 'Casablanca', 'MA', 'Morocco'),
    'Africa/Ceuta': ('Africa/Ceuta', 'Africa/Ceuta', 'Ceuta', 'ES', 'Spain'),
    'Africa/Conakry': ('Africa/Conakry', 'Africa/Conakry', 'Conakry', 'GN', 'Guinea'),
    'Africa/Dakar': ('Africa/Dakar', 'Africa/Dakar', 'Dakar', 'SN', 'Senegal'),
    'Africa/Dar_es_Salaam': ('Africa/Dar_es_Salaam', 'Africa/Dar es Salaam', 'Dar es Salaam', 'TZ', 'Tanzania'),
    'Africa/Djibouti': ('Africa/Djibouti', 'Africa/Djibouti', 'Djibouti', 'DJ', 'Djibouti'),
    'Africa/Douala': ('Africa/Douala', 'Africa/Douala', 'Douala', 'CM', 'Cameroon'),
    'Africa/El_Aaiun': ('Africa/El_Aaiun', 'Africa/El Aaiun', 'El Aaiun', 'EH', 'Western Sahara'),
    'Africa/Freetown': ('Africa/Freetown', 'Africa/Freetown', 'Freetown', 'SL', 'Sierra Leone'),
    'Africa/Gaborone': ('Africa/Gaborone', 'Africa/Gaborone', 'Gaborone', 'BW', 'Botswana'),
    'Africa/Harare': ('Africa/Harare', 'Africa/Harare', 'Harare', 'ZW', 'Zimbabwe'),
    'Africa/Johannesburg': ('Africa/Johannesburg', 'Africa/Johannesburg', 'Johannesburg', 'ZA', 'South Africa'),
    'Africa/Juba': ('Africa/Juba', 'Africa/Juba', 'Juba', 'SS', 'South Sudan'),
    'Africa/Kampala': ('Africa/Kampala', 'Africa/Kampala', 'Kampala', 'UG', 'Uganda'),
    'Africa/Khartoum': ('Africa/Khartoum', 'Africa/Khartoum', 'Khartoum', 'SD', 'Sudan'),
    'Africa/Kigali': ('Africa/Kigali', 'Africa/Kigali', 'Kigali', 'RW', 'Rwanda'),
    'Africa/Kinshasa': ('Africa/Kinshasa', 'Africa/Kinshasa', 'Kinshasa', 'CD', 'Congo (Dem. Rep.)'),
    'Africa/Lagos': ('Africa/Lagos', 'Africa/Lagos', 'Lagos', 'NG', 'Nigeria'),
    'Africa/Libreville': ('Africa/Libreville', 'Africa/Libreville', 'Libreville', 'GA', 'Gabon'),
    'Africa/Lome': ('Africa/Lome', 'Africa/Lome', 'Lome', 'TG', 'Togo'),
    'Africa/Luanda': ('Africa/Luanda', 'Africa/Luanda', 'Luanda', 'AO', 'Angola'),
    'Africa/Lubumbashi': ('Africa/Lubumbashi', 'Africa/Lubumbashi', 'Lubumbashi', 'CD', 'Congo (Dem. Rep.)'),
    'Africa/Lusaka': ('Africa/Lusaka', 'Africa/Lusaka', 'Lusaka', 'ZM', 'Zambia'),
    'Africa/Malabo': ('Africa/Malabo', 'Africa/Malabo', 'Malabo', 'GQ', 'Equatorial Guinea'),
    'Africa/Maputo': ('Africa/Maputo', 'Africa/Maputo', 'Maputo', 'MZ', 'Mozambique'),
    'Africa/Maseru': ('Africa/Maseru', 'Africa/Maseru', 'Maseru', 'LS', 'Lesotho'),
    'Africa/Mbabane': ('Africa/Mbabane', 'Africa/Mbabane', 'Mbabane', 'SZ', 'Eswatini (Swaziland)'),
    'Africa/Mogadishu': ('Africa/Mogadishu', 'Africa/Mogadishu', 'Mogadishu', 'SO', 'Somalia'),
    'Africa/Monrovia': ('Africa/Monrovia', 'Africa/Monrovia', 'Monrovia', 'LR', 'Liberia'),
    'Africa/Nairobi': ('Africa/Nairobi', 'Africa/Nairobi', 'Nairobi', 'KE', 'Kenya'),
    'Africa/Ndjamena': ('Africa/Ndjamena', 'Africa/Ndjamena', 'Ndjamena', 'TD', 'Chad'),
    'Africa/Niamey': ('Africa/Niamey', 'Africa/Niamey', 'Niamey', 'NE', 'Niger'),
    'Africa/Nouakchott': ('Africa/Nouakchott', 'Africa/Nouakchott', 'Nouakchott', 'MR', 'Mauritania'),
    'Africa/Ouagadougou': ('Africa/Ouagadougou', 'Africa/Ouagadougou', 'Ouagadougou', 'BF', 'Burkina Faso'),
    'Africa/Porto-Novo': ('Africa/Porto-Novo', 'Africa/Porto-Novo', 'Porto-Novo', 'BJ', 'Benin'),
    'Africa/Sao_Tome': ('Africa/Sao_Tome', 'Africa/Sao Tome', 'Sao Tome', 'ST', 'Sao Tome & Principe'),
    'Africa/Tripoli': ('Africa/Tripoli', 'Africa/Tripoli', 'Tripoli', 'LY', 'Libya'),
    'Africa/Tunis': ('Africa/Tunis', 'Africa/Tunis', 'Tunis', 'TN', 'Tunisia'),
    'Africa/Windhoek': ('Africa/Windhoek', 'Africa/Windhoek', 'Windhoek', 'NA', 'Namibia'),
    'America/Adak': ('America/Adak', 'America/Adak', 'Adak', 'US', 'United States'),
    'America/Anchorage': ('America/Anchorage', 'America/Anchorage', 'Anchorage', 'US', 'United States'),
    'America/Anguilla': ('America/Anguilla', 'America/Anguilla', 'Anguilla', 'AI', 'Anguilla'),
    'America/Antigua': ('America/Antigua', 'America/Antigua', 'Antigua', 'AG', 'Antigua & Barbuda'),
    'America/Araguaina': ('America/Araguaina', 'America/Araguaina', 'Araguaina', 'BR', 'Brazil'),
    'America/Argentina/Buenos_Aires': ('America/Argentina/Buenos_Aires', 'America/Argentina/Buenos Aires', 'Buenos Aires', 'AR', 'Argentina'),
    'America/Argentina/Catamarca': ('America/Argentina/Catamarca', 'America/Argentina/Catamarca', 'Catamarca', 'AR', 'Argentina'),
    'America/Argentina/Cordoba': ('America/Argentina/Cordoba', 'America/Argentina/Cordoba', 'Cordoba', 'AR', 'Argentina'),
    'America/Argentina/Jujuy': ('America/Argentina/Jujuy', 'America/Argentina/Jujuy', 'Jujuy', 'AR', 'Argentina'),
    'America/Argentina/La_Rioja': ('America/Argentina/La_Rioja', 'America/Argentina/La Rioja', 'La Rioja', 'AR', 'Argentina'),
    'America/Argentina/Mendoza': ('America/Argentina/Mendoza', 'America/Argentina/Mendoza', 'Mendoza', 'AR', 'Argentina'),
    'America/Argentina/Rio_Gallegos': ('America/Argentina/Rio_Gallegos', 'America/Argentina/Rio Gallegos', 'Rio Gallegos', 'AR', 'Argentina'),
    'America/Argentina/Salta': ('America/Argentina/Salta', 'America/Argentina/Salta', 'Salta', 'AR', 'Argentina'),
    'America/Argentina/San_Juan': ('America/Argentina/San_Juan', 'America/Argentina/San Juan', 'San Juan', 'AR', 'Argentina'),
    'America/Argentina/San_Luis': ('America/Argentina/San_Luis', 'America/Argentina/San Luis', 'San Luis', 'AR', 'Argentina'),
    'America/Argentina/Tucuman': ('America/Argentina/Tucuman', 'America/Argentina/Tucuman', 'Tucuman', 'AR', 'Argentina'),
    'America/Argentina/Ushuaia': ('America/Argentina/Ushuaia', 'America/Argentina/Ushuaia', 'Ushuaia', 'AR', 'Argentina'),
    'America/Aruba': ('America/Aruba', 'America/Aruba', 'Aruba', 'AW', 'Aruba'),
    'America/Asuncion': ('America/Asuncion', 'America/Asuncion', 'Asuncion', 'PY', 'Paraguay'),
    'America/Atikokan': ('America/Atikokan', 'America/Atikokan', 'Atikokan', 'CA', 'Canada'),
    'America/Bahia': ('America/Bahia', 'America/Bahia', 'Bahia', 'BR', 'Brazil'),
    'America/Bahia_Banderas': ('America/Bahia_Banderas', 'America/Bahia Banderas', 'Bahia Banderas', 'MX', 'Mexico'),
    'America/Barbados': ('America/Barbados', 'America/Barbados', 'Barbados', 'BB', 'Barbados'),
    'America/Belem': ('America/Belem', 'America/Belem', 'Belem', 'BR', 'Brazil'),
    'America/Belize': ('America/Belize', 'America/Belize', 'Belize', 'BZ', 'Belize'),
    'America/Blanc-Sablon': ('America/Blanc-Sablon', 'America/Blanc-Sablon', 'Blanc-Sablon', 'CA', 'Canada'),
    'America/Boa_Vista': ('America/Boa_Vista', 'America/Boa Vista', 'Boa Vista', 'BR', 'Brazil'),
    'America/Bogota': ('America/Bogota', 'America/Bogota', 'Bogota', 'CO', 'Colombia'),
    'America/Boise': ('America/Boise', 'America/Boise', 'Boise', 'US', 'United States'),
    'America/Cambridge_Bay': ('America/Cambridge_Bay', 'America/Cambridge Bay', 'Cambridge Bay', 'CA', 'Canada'),
    'America/Campo_Grande': ('America/Campo_Grande', 'America/Campo Grande', 'Campo Grande', 'BR', 'Brazil'),
    'America/Cancun': ('America/Cancun', 'America/Cancun', 'Cancun', 'MX', 'Mexico'),
    'America/Caracas': ('America/Caracas', 'America/Caracas', 'Caracas', 'VE', 'Venezuela'),
    'America/Cayenne': ('America/Cayenne', 'America/Cayenne', 'Cayenne', 'GF', 'French Guiana'),
    'America/Cayman': ('America/Cayman', 'America/Cayman', 'Cayman', 'KY', 'Cayman Islands'),
    'America/Chicago': ('America/Chicago', 'America/Chicago', 'Chicago', 'US', 'United States'),
    'America/Chihuahua': ('America/Chihuahua', 'America/Chihuahua', 'Chihuahua', 'MX', 'Mexico'),
    'America/Ciudad_Juarez': ('America/Ciudad_Juarez', 'America/Ciudad Juarez', 'Ciudad Juarez', 'MX', 'Mexico'),
    'America/Costa_Rica': ('America/Costa_Rica', 'America/Costa Rica', 'Costa Rica', 'CR', 'Costa Rica'),
    'America/Creston': ('America/Creston', 'America/Creston', 'Creston', 'CA', 'Canada'),
    'America/Cuiaba': ('America/Cuiaba', 'America/Cuiaba', 'Cuiaba', 'BR', 'Brazil'),
    'America/Curacao': ('America/Curacao', 'America/Curacao', 'Curacao', 'CW', 'Curaçao'),
    'America/Danmarkshavn': ('America/Danmarkshavn', 'America/Danmarkshavn', 'Danmarkshavn', 'GL', 'Greenland'),
    'America/Dawson': ('America/Dawson', 'America/Dawson', 'Dawson', 'CA', 'Canada'),
    'America/Dawson_Creek': ('America/Dawson_Creek', 'America/Dawson Creek', 'Dawson Creek', 'CA', 'Canada'),
    'America/Denver': ('America/Denver', 'America/Denver', 'Denver', 'US', 'United States'),
    'America/Detroit': ('America/Detroit', 'America/Detroit', 'Detroit', 'US', 'United States'),
    'America/Dominica': ('America/Dominica', 'America/Dominica', 'Dominica', 'DM', 'Dominica'),
    'America/Edmonton': ('America/Edmonton', 'America/Edmonton', 'Edmonton', 'CA', 'Canada'),
    'America/Eirunepe': ('America/Eirunepe', 'America/Eirunepe', 'Eirunepe', 'BR', 'Brazil'),
    'America/El_Salvador': ('America/El_Salvador', 'America/El Salvador', 'El Salvador', 'SV', 'El Salvador'),
    'America/Fort_Nelson': ('America/Fort_Nelson', 'America/Fort Nelson', 'Fort Nelson', 'CA', 'Canada'),
    'America/Fortaleza': ('America/Fortaleza', 'America/Fortaleza', 'Fortaleza', 'BR', 'Brazil'),
    'America/Glace_Bay': ('America/Glace_Bay', 'America/Glace Bay', 'Glace Bay', 'CA', 'Canada'),
    'America/Goose_Bay': ('America/Goose_Bay', 'America/Goose Bay', 'Goose Bay', 'CA', 'Canada'),
    'America/Grand_Turk': ('America/Grand_Turk', 'America/Grand Turk', 'Grand Turk', 'TC', 'Turks & Caicos Is'),
    'America/Grenada': ('America/Grenada', 'America/Grenada', 'Grenada', 'GD', 'Grenada'),
    'America/Guadeloupe': ('America/Guadeloupe', 'America/Guadeloupe', 'Guadeloupe', 'GP', 'Guadeloupe'),
    'America/Guatemala': ('America/Guatemala', 'America/Guatemala', 'Guatemala', 'GT', 'Guatemala'),
    'America/Guayaquil': ('America/Guayaquil', 'America/Guayaquil', 'Guayaquil', 'EC', 'Ecuador'),
    'America/Guyana': ('America/Guyana', 'America/Guyana', 'Guyana', 'GY', 'Guyana'),
    'America/Halifax': ('America/Halifax', 'America/Halifax', 'Halifax', 'CA', 'Canada'),
    'America/Havana': ('America/Havana', 'America/Havana', 'Havana', 'CU', 'Cuba'),
    'America/Hermosillo': ('America/Hermosillo', 'America/Hermosillo', 'Hermosillo', 'MX', 'Mexico'),
    'America/Indiana/Indianapolis': ('America/Indiana/Indianapolis', 'America/Indiana/Indianapolis', 'Indianapolis', 'US', 'United States'),
    'America/Indiana/Knox': ('America/Indiana/Knox', 'America/Indiana/Knox', 'Knox', 'US', 'United States'),
    'America/Indiana/Marengo': ('America/Indiana/Marengo', 'America/Indiana/Marengo', 'Marengo', 'US', 'United States'),
    'America/Indiana/Petersburg': ('America/Indiana/Petersburg', 'America/Indiana/Petersburg', 'Petersburg', 'US', 'United States'),
    'America/Indiana/Tell_City': ('America/Indiana/Tell_City', 'America/Indiana/Tell City', 'Tell City', 'US', 'United States'),
    'America/Indiana/Vevay': ('America/Indiana/Vevay', 'America/Indiana/Vevay', 'Vevay', 'US', 'United States'),
    'America/Indiana/Vincennes': ('America/Indiana/Vincennes', 'America/Indiana/Vincennes', 'Vincennes', 'US', 'United States'),
    'America/Indiana/Winamac': ('America/Indiana/Winamac', 'America/Indiana/Winamac', 'Winamac', 'US', 'United States'),
    'America/Inuvik': ('America/Inuvik', 'America/Inuvik', 'Inuvik', 'CA', 'Canada'),
    'America/Iqaluit': ('America/Iqaluit', 'America/Iqaluit', 'Iqaluit', 'CA', 'Canada'),
    'America/Jamaica': ('America/Jamaica', 'America/Jamaica', 'Jamaica', 'JM', 'Jamaica'),
    'America/Juneau': ('America/Juneau', 'America/Juneau', 'Juneau', 'US', 'United States'),
    'America/Kentucky/Louisville': ('America/Kentucky/Louisville', 'America/Kentucky/Louisville', 'Louisville', 'US', 'United States'),
    'America/Kentucky/Monticello': ('America/Kentucky/Monticello', 'America/Kentucky/Monticello', 'Monticello', 'US', 'United States'),
    'America/Kralendijk': ('America/Kralendijk', 'America/Kralendijk', 'Kralendijk', 'BQ', 'Caribbean NL'),
    'America/La_Paz': ('America/La_Paz', 'America/La Paz', 'La Paz', 'BO', 'Bolivia'),
    'America/Lima': ('America/Lima', 'America/Lima', 'Lima', 'PE', 'Peru'),
    'America/Los_Angeles': ('America/Los_Angeles', 'America/Los Angeles', 'Los Angeles', 'US', 'United States'),
    'America/Lower_Princes': ('America/Lower_Princes', 'America/Lower Princes', 'Lower Princes', 'SX', 'St Maarten (Dutch)'),
    'America/Maceio': ('America/Maceio', 'America/Maceio', 'Maceio', 'BR', 'Brazil'),
    'America/Managua': ('America/Managua', 'America/Managua', 'Managua', 'NI', 'Nicaragua'),
    'America/Manaus': ('America/Manaus', 'America/Manaus', 'Manaus', 'BR', 'Brazil'),
    'America/Marigot': ('America/Marigot', 'America/Marigot', 'Marigot', 'MF', 'St Martin (French)'),
    'America/Martinique': ('America/Martinique', 'America/Martinique', 'Martinique', 'MQ', 'Martinique'),
    'America/Matamoros': ('America/Matamoros', 'America/Matamoros', 'Matamoros', 'MX', 'Mexico'),
    'America/Mazatlan': ('America/Mazatlan', 'America/Mazatlan', 'Mazatlan', 'MX', 'Mexico'),
    'America/Menominee': ('America/Menominee', 'America/Menominee', 'Menominee', 'US', 'United States'),
    'America/Merida': ('America/Merida', 'America/Merida', 'Merida', 'MX', 'Mexico'),
    'America/Metlakatla': ('America/Metlakatla', 'America/Metlakatla', 'Metlakatla', 'US', 'United States'),
    'America/Mexico_City': ('America/Mexico_City', 'America/Mexico City', 'Mexico City', 'MX', 'Mexico'),
    'America/Miquelon': ('America/Miquelon', 'America/Miquelon', 'Miquelon', 'PM', 'St Pierre & Miquelon'),
    'America/Moncton': ('America/Moncton', 'America/Moncton', 'Moncton', 'CA', 'Canada'),
    'America/Monterrey': ('America/Monterrey', 'America/Monterrey', 'Monterrey', 'MX', 'Mexico'),
    'America/Montevideo': ('America/Montevideo', 'America/Montevideo', 'Montevideo', 'UY', 'Uruguay'),
    'America/Montserrat': ('America/Montserrat', 'America/Montserrat', 'Montserrat', 'MS', 'Montserrat'),
    'America/Nassau': ('America/Nassau', 'America/Nassau', 'Nassau', 'BS', 'Bahamas'),
    'America/New_York': ('America/New_York', 'America/New York', 'New York', 'US', 'United States'),
    'America/Nome': ('America/Nome', 'America/Nome', 'Nome', 'US', 'United States'),
    'America/Noronha': ('America/Noronha', 'America/Noronha', 'Noronha', 'BR', 'Brazil'),
    'America/North_Dakota/Beulah': ('America/North_Dakota/Beulah', 'America/North Dakota/Beulah', 'Beulah', 'US', 'United States'),
    'America/North_Dakota/Center': ('America/North_Dakota/Center', 'America/North Dakota/Center', 'Center', 'US', 'United States'),
    'America/North_Dakota/New_Salem': ('America/North_Dakota/New_Salem', 'America/North Dakota/New Salem', 'New Salem', 'US', 'United States'),
    'America/Nuuk': ('America/Nuuk', 'America/Nuuk', 'Nuuk', 'GL', 'Greenland'),
    'America/Ojinaga': ('America/Ojinaga', 'America/Ojinaga', 'Ojinaga', 'MX', 'Mexico'),
    'America/Panama': ('America/Panama', 'America/Panama', 'Panama', 'PA', 'Panama'),
    'America/Paramaribo': ('America/Paramaribo', 'America/Paramaribo', 'Paramaribo', 'SR', 'Suriname'),
    'America/Phoenix': ('America/Phoenix', 'America/Phoenix', 'Phoenix', 'US', 'United States'),
    'America/Port-au-Prince': ('America/Port-au-Prince', 'America/Port-au-Prince', 'Port-au-Prince', 'HT', 'Haiti'),
    'America/Port_of_Spain': ('America/Port_of_Spain', 'America/Port of Spain', 'Port of Spain', 'TT', 'Trinidad & Tobago'),
    'America/Porto_Velho': ('America/Porto_Velho', 'America/Porto Velho', 'Porto Velho', 'BR', 'Brazil'),
    'America/Puerto_Rico': ('America/Puerto_Rico', 'America/Puerto Rico', 'Puerto Rico', 'PR', 'Puerto Rico'),
    'America/Punta_Arenas': ('America/Punta_Arenas', 'America/Punta Arenas', 'Punta Arenas', 'CL', 'Chile'),
    'America/Rankin_Inlet': ('America/Rankin_Inlet', 'America/Rankin Inlet', 'Rankin Inlet', 'CA', 'Canada'),
    'America/Recife': ('America/Recife', 'America/Recife', 'Recife', 'BR', 'Brazil'),
    'America/Regina': ('America/Regina', 'America/Regina', 'Regina', 'CA', 'Canada'),
    'America/Resolute': ('America/Resolute', 'America/Resolute', 'Resolute', 'CA', 'Canada'),
    'America/Rio_Branco': ('America/Rio_Branco', 'America/Rio Branco', 'Rio Branco', 'BR', 'Brazil'),
    'America/Santarem': ('America/Santarem', 'America/Santarem', 'Santarem', 'BR', 'Brazil'),
    'America/Santiago': ('America/Santiago', 'America/Santiago', 'Santiago', 'CL', 'Chile'),
    'America/Santo_Domingo': ('America/Santo_Domingo', 'America/Santo Domingo', 'Santo Domingo', 'DO', 'Dominican Republic'),
    'America/Sao_Paulo': ('America/Sao_Paulo', 'America/Sao Paulo', 'Sao Paulo', 'BR', 'Brazil'),
    'America/Scoresbysund': ('America/Scoresbysund', 'America/Scoresbysund', 'Scoresbysund', 'GL', 'Greenland'),
    'America/Sitka': ('America/Sitka', 'America/Sitka', 'Sitka', 'US', 'United States'),
    'America/St_Barthelemy': ('America/St_Barthelemy', 'America/St Barthelemy', 'St Barthelemy', 'BL', 'St Barthelemy'),
    'America/St_Johns': ('America/St_Johns', 'America/St Johns', 'St Johns', 'CA', 'Canada'),
    'America/St_Kitts': ('America/St_Kitts', 'America/St Kitts', 'St Kitts', 'KN', 'St Kitts & Nevis'),
    'America/St_Lucia': ('America/St_Lucia', 'America/St Lucia', 'St Lucia', 'LC', 'St Lucia'),
    'America/St_Thomas': ('America/St_Thomas', 'America/St Thomas', 'St Thomas', 'VI', 'Virgin Islands (US)'),
    'America/St_Vincent': ('America/St_Vincent', 'America/St Vincent', 'St Vincent', 'VC', 'St Vincent'),
    'America/Swift_Current': ('America/Swift_Current', 'America/Swift Current', 'Swift Current', 'CA', 'Canada'),
    'America/Tegucigalpa': ('America/Tegucigalpa', 'America/Tegucigalpa', 'Tegucigalpa', 'HN', 'Honduras'),
    'America/Thule': ('America/Thule', 'America/Thule', 'Thule', 'GL', 'Greenland'),
    'America/Tijuana': ('America/Tijuana', 'America/Tijuana', 'Tijuana', 'MX', 'Mexico'),
    'America/Toronto': ('America/Toronto', 'America/Toronto', 'Toronto', 'CA', 'Canada'),
    'America/Tortola': ('America/Tortola', 'America/Tortola', 'Tortola', 'VG', 'Virgin Islands (UK)'),
    'America/Vancouver': ('America/Vancouver', 'America/Vancouver', 'Vancouver', 'CA', 'Canada'),
    'America/Whitehorse': ('America/Whitehorse', 'America/Whitehorse', 'Whitehorse', 'CA', 'Canada'),
    'America/Winnipeg': ('America/Winnipeg', 'America/Winnipeg', 'Winnipeg', 'CA', 'Canada'),
    'America/Yakutat': ('America/Yakutat', 'America/Yakutat', 'Yakutat', 'US', 'United States'),
    'Antarctica/Casey': ('Antarctica/Casey', 'Antarctica/Casey', 'Casey', 'AQ', 'Antarctica'),
    'Antarctica/Davis': ('Antarctica/Davis', 'Antarctica/Davis', 'Davis', 'AQ', 'Antarctica'),
    'Antarctica/DumontDUrville': ('Antarctica/DumontDUrville', 'Antarctica/DumontDUrville', 'DumontDUrville', 'AQ', 'Antarctica'),
    'Antarctica/Macquarie': ('Antarctica/Macquarie', 'Antarctica/Macquarie', 'Macquarie', 'AU', 'Australia'),
    'Antarctica/Mawson': ('Antarctica/Mawson', 'Antarctica/Mawson', 'Mawson', 'AQ', 'Antarctica'),
    'Antarctica/McMurdo': ('Antarctica/McMurdo', 'Antarctica/McMurdo', 'McMurdo', 'AQ', 'Antarctica'),
    'Antarctica/Palmer': ('Antarctica/Palmer', 'Antarctica/Palmer', 'Palmer', 'AQ', 'Antarctica'),
    'Antarctica/Rothera': ('Antarctica/Rothera', 'Antarctica/Rothera', 'Rothera', 'AQ', 'Antarctica'),
    'Antarctica/Syowa': ('Antarctica/Syowa', 'Antarctica/Syowa', 'Syowa', 'AQ', 'Antarctica'),
    'Antarctica/Troll': ('Antarctica/Troll', 'Antarctica/Troll', 'Troll', 'AQ', 'Antarctica'),
    'Antarctica/Vostok': ('Antarctica/Vostok', 'Antarctica/Vostok', 'Vostok', 'AQ', 'Antarctica'),
    'Arctic/Longyearbyen': ('Arctic/Longyearbyen', 'Arctic/Longyearbyen', 'Longyearbyen', 'SJ', 'Svalbard & Jan Mayen'),
    'Asia/Aden': ('Asia/Aden', 'Asia/Aden', 'Aden', 'YE', 'Yemen'),
    'Asia/Almaty': ('Asia/Almaty', 'Asia/Almaty', 'Almaty', 'KZ', 'Kazakhstan'),
    'Asia/Amman': ('Asia/Amman', 'Asia/Amman', 'Amman', 'JO', 'Jordan'),
    'Asia/Anadyr': ('Asia/Anadyr', 'Asia/Anadyr', 'Anadyr', 'RU', 'Russia'),
    'Asia/Aqtau': ('Asia/Aqtau', 'Asia/Aqtau', 'Aqtau', 'KZ', 'Kazakhstan'),
    'Asia/Aqtobe': ('Asia/Aqtobe', 'Asia/Aqtobe', 'Aqtobe', 'KZ', 'Kazakhstan'),
    'Asia/Ashgabat': ('Asia/Ashgabat', 'Asia/Ashgabat', 'Ashgabat', 'TM', 'Turkmenistan'),
    'Asia/Atyrau': ('Asia/Atyrau', 'Asia/Atyrau', 'Atyrau', 'KZ', 'Kazakhstan'),
    'Asia/Baghdad': ('Asia/Baghdad', 'Asia/Baghdad', 'Baghdad', 'IQ', 'Iraq'),
    'Asia/Bahrain': ('Asia/Bahrain', 'Asia/Bahrain', 'Bahrain', 'BH', 'Bahrain'),
    'Asia/Baku': ('Asia/Baku', 'Asia/Baku', 'Baku', 'AZ', 'Azerbaijan'),
    'Asia/Bangkok': ('Asia/Bangkok', 'Asia/Bangkok', 'Bangkok', 'TH', 'Thailand'),
    'Asia/Barnaul': ('Asia/Barnaul', 'Asia/Barnaul', 'Barnaul', 'RU', 'Russia'),
    'Asia/Beirut': ('Asia/Beirut', 'Asia/Beirut', 'Beirut', 'LB', 'Lebanon'),
    'Asia/Bishkek': ('Asia/Bishkek', 'Asia/Bishkek', 'Bishkek', 'KG', 'Kyrgyzstan'),
    'Asia/Brunei': ('Asia/Brunei', 'Asia/Brunei', 'Brunei', 'BN', 'Brunei'),
    'Asia/Chita': ('Asia/Chita', 'Asia/Chita', 'Chita', 'RU', 'Russia'),
    'Asia/Colombo': ('Asia/Colombo', 'Asia/Colombo', 'Colombo', 'LK', 'Sri Lanka'),
    'Asia/Damascus': ('Asia/Damascus', 'Asia/Damascus', 'Damascus', 'SY', 'Syria'),
    'Asia/Dhaka': ('Asia/Dhaka', 'Asia/Dhaka', 'Dhaka', 'BD', 'Bangladesh'),
    'Asia/Dili': ('Asia/Dili', 'Asia/Dili', 'Dili', 'TL', 'East Timor'),
    'Asia/Dubai': ('Asia/Dubai', 'Asia/Dubai', 'Dubai', 'AE', 'United Arab Emirates'),
    'Asia/Dushanbe': ('Asia/Dushanbe', 'Asia/Dushanbe', 'Dushanbe', 'TJ', 'Tajikistan'),
    'Asia/Famagusta': ('Asia/Famagusta', 'Asia/Famagusta', 'Famagusta', 'CY', 'Cyprus'),
    'Asia/Gaza': ('Asia/Gaza', 'Asia/Gaza', 'Gaza', 'PS', 'Palestine'),
    'Asia/Hebron': ('Asia/Hebron', 'Asia/Hebron', 'Hebron', 'PS', 'Palestine'),
    'Asia/Ho_Chi_Minh': ('Asia/Ho_Chi_Minh', 'Asia/Ho Chi Minh', 'Ho Chi Minh', 'VN', 'Vietnam'),
    'Asia/Hong_Kong': ('Asia/Hong_Kong', 'Asia/Hong Kong', 'Hong Kong', 'HK', 'Hong Kong'),
    'Asia/Hovd': ('Asia/Hovd', 'Asia/Hovd', 'Hovd', 'MN', 'Mongolia'),
    'Asia/Irkutsk': ('Asia/Irkutsk', 'Asia/Irkutsk', 'Irkutsk', 'RU', 'Russia'),
    'Asia/Jakarta': ('Asia/Jakarta', 'Asia/Jakarta', 'Jakarta', 'ID', 'Indonesia'),
    'Asia/Jayapura': ('Asia/Jayapura', 'Asia/Jayapura', 'Jayapura', 'ID', 'Indonesia'),
    'Asia/Jerusalem': ('Asia/Jerusalem', 'Asia/Jerusalem', 'Jerusalem', 'IL', 'Israel'),
    'Asia/Kabul': ('Asia/Kabul', 'Asia/Kabul', 'Kabul', 'AF', 'Afghanistan'),
    'Asia/Kamchatka': ('Asia/Kamchatka', 'Asia/Kamchatka', 'Kamchatka', 'RU', 'Russia'),
    'Asia/Karachi': ('Asia/Karachi', 'Asia/Karachi', 'Karachi', 'PK', 'Pakistan'),
    'Asia/Kathmandu': ('Asia/Kathmandu', 'Asia/Kathmandu', 'Kathmandu', 'NP', 'Nepal'),
    'Asia/Khandyga': ('Asia/Khandyga', 'Asia/Khandyga', 'Khandyga', 'RU', 'Russia'),
    'Asia/Kolkata': ('Asia/Kolkata', 'Asia/Kolkata', 'Kolkata', 'IN', 'India'),
    'Asia/Krasnoyarsk': ('Asia/Krasnoyarsk', 'Asia/Krasnoyarsk', 'Krasnoyarsk', 'RU', 'Russia'),
    'Asia/Kuala_Lumpur': ('Asia/Kuala_Lumpur', 'Asia/Kuala Lumpur', 'Kuala Lumpur', 'MY', 'Malaysia'),
    'Asia/Kuching': ('Asia/Kuching', 'Asia/Kuching', 'Kuching', 'MY', 'Malaysia'),
    'Asia/Kuwait': ('Asia/Kuwait', 'Asia/Kuwait', 'Kuwait', 'KW', 'Kuwait'),
    'Asia/Macau': ('Asia/Macau', 'Asia/Macau', 'Macau', 'MO', 'Macau'),
    'Asia/Magadan': ('Asia/Magadan', 'Asia/Magadan', 'Magadan', 'RU', 'Russia'),
    'Asia/Makassar': ('Asia/Makassar', 'Asia/Makassar', 'Makassar', 'ID', 'Indonesia'),
    'Asia/Manila': ('Asia/Manila', 'Asia/Manila', 'Manila', 'PH', 'Philippines'),
    'Asia/Muscat': ('Asia/Muscat', 'Asia/Muscat', 'Muscat', 'OM', 'Oman'),
    'Asia/Nicosia': ('Asia/Nicosia', 'Asia/Nicosia', 'Nicosia', 'CY', 'Cyprus'),
    'Asia/Novokuznetsk': ('Asia/Novokuznetsk', 'Asia/Novokuznetsk', 'Novokuznetsk', 'RU', 'Russia'),
    'Asia/Novosibirsk': ('Asia/Novosibirsk', 'Asia/Novosibirsk', 'Novosibirsk', 'RU', 'Russia'),
    'Asia/Omsk': ('Asia/Omsk', 'Asia/Omsk', 'Omsk', 'RU', 'Russia'),
    'Asia/Oral': ('Asia/Oral', 'Asia/Oral', 'Oral', 'KZ', 'Kazakhstan'),
    'Asia/Phnom_Penh': ('Asia/Phnom_Penh', 'Asia/Phnom Penh', 'Phnom Penh', 'KH', 'Cambodia'),
    'Asia/Pontianak': ('Asia/Pontianak', 'Asia/Pontianak', 'Pontianak', 'ID', 'Indonesia'),
    'Asia/Pyongyang': ('Asia/Pyongyang', 'Asia/Pyongyang', 'Pyongyang', 'KP', 'Korea (North)'),
    'Asia/Qatar': ('Asia/Qatar', 'Asia/Qatar', 'Qatar', 'QA', 'Qatar'),
    'Asia/Qostanay': ('Asia/Qostanay', 'Asia/Qostanay', 'Qostanay', 'KZ', 'Kazakhstan'),
    'Asia/Qyzylorda': ('Asia/Qyzylorda', 'Asia/Qyzylorda', 'Qyzylorda', 'KZ', 'Kazakhstan'),
    'Asia/Riyadh': ('Asia/Riyadh', 'Asia/Riyadh', 'Riyadh', 'SA', 'Saudi Arabia'),
    'Asia/Sakhalin': ('Asia/Sakhalin', 'Asia/Sakhalin', 'Sakhalin', 'RU', 'Russia'),
    'Asia/Samarkand': ('Asia/Samarkand', 'Asia/Samarkand', 'Samarkand', 'UZ', 'Uzbekistan'),
    'Asia/Seoul': ('Asia/Seoul', 'Asia/Seoul', 'Seoul', 'KR', 'Korea (South)'),
    'Asia/Shanghai': ('Asia/Shanghai', 'Asia/Shanghai', 'Shanghai', 'CN', 'China'),
    'Asia/Singapore': ('Asia/Singapore', 'Asia/Singapore', 'Singapore', 'SG', 'Singapore'),
    'Asia/Srednekolymsk': ('Asia/Srednekolymsk', 'Asia/Srednekolymsk', 'Srednekolymsk', 'RU', 'Russia'),
    'Asia/Taipei': ('Asia/Taipei', 'Asia/Taipei', 'Taipei', 'TW', 'Taiwan'),
    'Asia/Tashkent': ('Asia/Tashkent', 'Asia/Tashkent', 'Tashkent', 'UZ', 'Uzbekistan'),
    'Asia/Tbilisi': ('Asia/Tbilisi', 'Asia/Tbilisi', 'Tbilisi', 'GE', 'Georgia'),
    'Asia/Tehran': ('Asia/Tehran', 'Asia/Tehran', 'Tehran', 'IR', 'Iran'),
    'Asia/Thimphu': ('Asia/Thimphu', 'Asia/Thimphu', 'Thimphu', 'BT', 'Bhutan'),
    'Asia/Tokyo': ('Asia/Tokyo', 'Asia/Tokyo', 'Tokyo', 'JP', 'Japan'),
    'Asia/Tomsk': ('Asia/Tomsk', 'Asia/Tomsk', 'Tomsk', 'RU', 'Russia'),
    'Asia/Ulaanbaatar': ('Asia/Ulaanbaatar', 'Asia/Ulaanbaatar', 'Ulaanbaatar', 'MN', 'Mongolia'),
    'Asia/Urumqi': ('Asia/Urumqi', 'Asia/Urumqi', 'Urumqi', 'CN', 'China'),
    'Asia/Ust-Nera': ('Asia/Ust-Nera', 'Asia/Ust-Nera', 'Ust-Nera', 'RU', 'Russia'),
    'Asia/Vientiane': ('Asia/Vientiane', 'Asia/Vientiane', 'Vientiane', 'LA', 'Laos'),
    'Asia/Vladivostok': ('Asia/Vladivostok', 'Asia/Vladivostok', 'Vladivostok', 'RU', 'Russia'),
    'Asia/Yakutsk': ('Asia/Yakutsk', 'Asia/Yakutsk', 'Yakutsk', 'RU', 'Russia'),
    'Asia/Yangon': ('Asia/Yangon', 'Asia/Yangon', 'Yangon', 'MM', 'Myanmar (Burma)'),
    'Asia/Yekaterinburg': ('Asia/Yekaterinburg', 'Asia/Yekaterinburg', 'Yekaterinburg', 'RU', 'Russia'),
    'Asia/Yerevan': ('Asia/Yerevan', 'Asia/Yerevan', 'Yerevan', 'AM', 'Armenia'),
    'Atlantic/Azores': ('Atlantic/Azores', 'Atlantic/Azores', 'Azores', 'PT', 'Portugal'),
    'Atlantic/Bermuda': ('Atlantic/Bermuda', 'Atlantic/Bermuda', 'Bermuda', 'BM', 'Bermuda'),
    'Atlantic/Canary': ('Atlantic/Canary', 'Atlantic/Canary', 'Canary', 'ES', 'Spain'),
    'Atlantic/Cape_Verde': ('Atlantic/Cape_Verde', 'Atlantic/Cape Verde', 'Cape Verde', 'CV', 'Cape Verde'),
    'Atlantic/Faroe': ('Atlantic/Faroe', 'Atlantic/Faroe', 'Faroe', 'FO', 'Faroe Islands'),
    'Atlantic/Madeira': ('Atlantic/Madeira', 'Atlantic/Madeira', 'Madeira', 'PT', 'Portugal'),
    'Atlantic/Reykjavik': ('Atlantic/Reykjavik', 'Atlantic/Reykjavik', 'Reykjavik', 'IS', 'Iceland'),
    'Atlantic/South_Georgia': ('Atlantic/South_Georgia', 'Atlantic/South Georgia', 'South Georgia', 'GS', 'South Georgia & the South Sandwich Islands'),
    'Atlantic/St_Helena': ('Atlantic/St_Helena', 'Atlantic/St Helena', 'St Helena', 'SH', 'St Helena'),
    'Atlantic/Stanley': ('Atlantic/Stanley', 'Atlantic/Stanley', 'Stanley', 'FK', 'Falkland Islands'),
    'Australia/Adelaide': ('Australia/Adelaide', 'Australia/Adelaide', 'Adelaide', 'AU', 'Australia'),
    'Australia/Brisbane': ('Australia/Brisbane', 'Australia/Brisbane', 'Brisbane', 'AU', 'Australia'),
    'Australia/Broken_Hill': ('Australia/Broken_Hill', 'Australia/Broken Hill', 'Broken Hill', 'AU', 'Australia'),
    'Australia/Darwin': ('Australia/Darwin', 'Australia/Darwin', 'Darwin', 'AU', 'Australia'),
    'Australia/Eucla': ('Australia/Eucla', 'Australia/Eucla', 'Eucla', 'AU', 'Australia'),
    'Australia/Hobart': ('Australia/Hobart', 'Australia/Hobart', 'Hobart', 'AU', 'Australia'),
    'Australia/Lindeman': ('Australia/Lindeman', 'Australia/Lindeman', 'Lindeman', 'AU', 'Australia'),
    'Australia/Lord_Howe': ('Australia/Lord_Howe', 'Australia/Lord Howe', 'Lord Howe', 'AU', 'Australia'),
    'Australia/Melbourne': ('Australia/Melbourne', 'Australia/Melbourne', 'Melbourne', 'AU', 'Australia'),
    'Australia/Perth': ('Australia/Perth', 'Australia/Perth', 'Perth', 'AU', 'Australia'),
    'Australia/Sydney': ('Australia/Sydney', 'Australia/Sydney', 'Sydney', 'AU', 'Australia'),
    'Europe/Amsterdam': ('Europe/Amsterdam', 'Europe/Amsterdam', 'Amsterdam', 'NL', 'Netherlands'),
    'Europe/Andorra': ('Europe/Andorra', 'Europe/Andorra', 'Andorra', 'AD', 'Andorra'),
    'Europe/Astrakhan': ('Europe/Astrakhan', 'Europe/Astrakhan', 'Astrakhan', 'RU', 'Russia'),
    'Europe/Athens': ('Europe/Athens', 'Europe/Athens', 'Athens', 'GR', 'Greece'),
    'Europe/Belgrade': ('Europe/Belgrade', 'Europe/Belgrade', 'Belgrade', 'RS', 'Serbia'),
    'Europe/Berlin': ('Europe/Berlin', 'Europe/Berlin', 'Berlin', 'DE', 'Germany'),
    'Europe/Bratislava': ('Europe/Bratislava', 'Europe/Bratislava', 'Bratislava', 'SK', 'Slovakia'),
    'Europe/Brussels': ('Europe/Brussels', 'Europe/Brussels', 'Brussels', 'BE', 'Belgium'),
    'Europe/Bucharest': ('Europe/Bucharest', 'Europe/Bucharest', 'Bucharest', 'RO', 'Romania'),
    'Europe/Budapest': ('Europe/Budapest', 'Europe/Budapest', 'Budapest', 'HU', 'Hungary'),
    'Europe/Busingen': ('Europe/Busingen', 'Europe/Busingen', 'Busingen', 'DE', 'Germany'),
    'Europe/Chisinau': ('Europe/Chisinau', 'Europe/Chisinau', 'Chisinau', 'MD', 'Moldova'),
    'Europe/Copenhagen': ('Europe/Copenhagen', 'Europe/Copenhagen', 'Copenhagen', 'DK', 'Denmark'),
    'Europe/Dublin': ('Europe/Dublin', 'Europe/Dublin', 'Dublin', 'IE', 'Ireland'),
    'Europe/Gibraltar': ('Europe/Gibraltar', 'Europe/Gibraltar', 'Gibraltar', 'GI', 'Gibraltar'),
    'Europe/Guernsey': ('Europe/Guernsey', 'Europe/Guernsey', 'Guernsey', 'GG', 'Guernsey'),
    'Europe/Helsinki': ('Europe/Helsinki', 'Europe/Helsinki', 'Helsinki', 'FI', 'Finland'),
    'Europe/Isle_of_Man': ('Europe/Isle_of_Man', 'Europe/Isle of Man', 'Isle of Man', 'IM', 'Isle of Man'),
    'Europe/Istanbul': ('Europe/Istanbul', 'Europe/Istanbul', 'Istanbul', 'TR', 'Turkey'),
    'Europe/Jersey': ('Europe/Jersey', 'Europe/Jersey', 'Jersey', 'JE', 'Jersey'),
    'Europe/Kaliningrad': ('Europe/Kaliningrad', 'Europe/Kaliningrad', 'Kaliningrad', 'RU', 'Russia'),
    'Europe/Kirov': ('Europe/Kirov', 'Europe/Kirov', 'Kirov', 'RU', 'Russia'),
    'Europe/Kyiv': ('Europe/Kyiv', 'Europe/Kyiv', 'Kyiv', 'UA', 'Ukraine'),
    'Europe/Lisbon': ('Europe/Lisbon', 'Europe/Lisbon', 'Lisbon', 'PT', 'Portugal'),
    'Europe/Ljubljana': ('Europe/Ljubljana', 'Europe/Ljubljana', 'Ljubljana', 'SI', 'Slovenia'),
    'Europe/London': ('Europe/London', 'Europe/London', 'London', 'GB', 'Britain (UK)'),
    'Europe/Luxembourg': ('Europe/Luxembourg', 'Europe/Luxembourg', 'Luxembourg', 'LU', 'Luxembourg'),
    'Europe/Madrid': ('Europe/Madrid', 'Europe/Madrid', 'Madrid', 'ES', 'Spain'),
    'Europe/Malta': ('Europe/Malta', 'Europe/Malta', 'Malta', 'MT', 'Malta'),
    'Europe/Mariehamn': ('Europe/Mariehamn', 'Europe/Mariehamn', 'Mariehamn', 'AX', 'Åland Islands'),
    'Europe/Minsk': ('Europe/Minsk', 'Europe/Minsk', 'Minsk', 'BY', 'Belarus'),
    'Europe/Monaco': ('Europe/Monaco', 'Europe/Monaco', 'Monaco', 'MC', 'Monaco'),
    'Europe/Moscow': ('Europe/Moscow', 'Europe/Moscow', 'Moscow', 'RU', 'Russia'),
    'Europe/Oslo': ('Europe/Oslo', 'Europe/Oslo', 'Oslo', 'NO', 'Norway'),
    'Europe/Paris': ('Europe/Paris', 'Europe/Paris', 'Paris', 'FR', 'France'),
    'Europe/Podgorica': ('Europe/Podgorica', 'Europe/Podgorica', 'Podgorica', 'ME', 'Montenegro'),
    'Europe/Prague': ('Europe/Prague', 'Europe/Prague', 'Prague', 'CZ', 'Czech Republic'),
    'Europe/Riga': ('Europe/Riga', 'Europe/Riga', 'Riga', 'LV', 'Latvia'),
    'Europe/Rome': ('Europe/Rome', 'Europe/Rome', 'Rome', 'IT', 'Italy'),
    'Europe/Samara': ('Europe/Samara', 'Europe/Samara', 'Samara', 'RU', 'Russia'),
    'Europe/San_Marino': ('Europe/San_Marino', 'Europe/San Marino', 'San Marino', 'SM', 'San Marino'),
    'Europe/Sarajevo': ('Europe/Sarajevo', 'Europe/Sarajevo', 'Sarajevo', 'BA', 'Bosnia & Herzegovina'),
    'Europe/Saratov': ('Europe/Saratov', 'Europe/Saratov', 'Saratov', 'RU', 'Russia'),
    'Europe/Simferopol': ('Europe/Simferopol', 'Europe/Simferopol', 'Simferopol', 'UA', 'Ukraine'),
    'Europe/Skopje': ('Europe/Skopje', 'Europe/Skopje', 'Skopje', 'MK', 'North Macedonia'),
    'Europe/Sofia': ('Europe/Sofia', 'Europe/Sofia', 'Sofia', 'BG', 'Bulgaria'),
    'Europe/Stockholm': ('Europe/Stockholm', 'Europe/Stockholm', 'Stockholm', 'SE', 'Sweden'),
    'Europe/Tallinn': ('Europe/Tallinn', 'Europe/Tallinn', 'Tallinn', 'EE', 'Estonia'),
    'Europe/Tirane': ('Europe/Tirane', 'Europe/Tirane', 'Tirane', 'AL', 'Albania'),
    'Europe/Ulyanovsk': ('Europe/Ulyanovsk', 'Europe/Ulyanovsk', 'Ulyanovsk', 'RU', 'Russia'),
    'Europe/Vaduz': ('Europe/Vaduz', 'Europe/Vaduz', 'Vaduz', 'LI', 'Liechtenstein'),
    'Europe/Vatican': ('Europe/Vatican', 'Europe/Vatican', 'Vatican', 'VA', 'Vatican City'),
    'Europe/Vienna': ('Europe/Vienna', 'Europe/Vienna', 'Vienna', 'AT', 'Austria'),
    'Europe/Vilnius': ('Europe/Vilnius', 'Europe/Vilnius', 'Vilnius', 'LT', 'Lithuania'),
    'Europe/Volgograd': ('Europe/Volgograd', 'Europe/Volgograd', 'Volgograd', 'RU', 'Russia'),
    'Europe/Warsaw': ('Europe/Warsaw', 'Europe/Warsaw', 'Warsaw', 'PL', 'Poland'),
    'Europe/Zagreb': ('Europe/Zagreb', 'Europe/Zagreb', 'Zagreb', 'HR', 'Croatia'),
    'Europe/Zurich': ('Europe/Zurich', 'Europe/Zurich', 'Zurich', 'CH', 'Switzerland'),
    'Indian/Antananarivo': ('Indian/Antananarivo', 'Indian/Antananarivo', 'Antananarivo', 'MG', 'Madagascar'),
    'Indian/Chagos': ('Indian/Chagos', 'Indian/Chagos', 'Chagos', 'IO', 'British Indian Ocean Territory'),
    'Indian/Christmas': ('Indian/Christmas', 'Indian/Christmas', 'Christmas', 'CX', 'Christmas Island'),
    'Indian/Cocos': ('Indian/Cocos', 'Indian/Cocos', 'Cocos', 'CC', 'Cocos (Keeling) Islands'),
    'Indian/Comoro': ('Indian/Comoro', 'Indian/Comoro', 'Comoro', 'KM', 'Comoros'),
    'Indian/Kerguelen': ('Indian/Kerguelen', 'Indian/Kerguelen', 'Kerguelen', 'TF', 'French S. Terr.'),
    'Indian/Mahe': ('Indian/Mahe', 'Indian/Mahe', 'Mahe', 'SC', 'Seychelles'),
    'Indian/Maldives': ('Indian/Maldives', 'Indian/Maldives', 'Maldives', 'MV', 'Maldives'),
    'Indian/Mauritius': ('Indian/Mauritius', 'Indian/Mauritius', 'Mauritius', 'MU', 'Mauritius'),
    'Indian/Mayotte': ('Indian/Mayotte', 'Indian/Mayotte', 'Mayotte', 'YT', 'Mayotte'),
    'Indian/Reunion': ('Indian/Reunion', 'Indian/Reunion', 'Reunion', 'RE', 'Réunion'),
    'Pacific/Apia': ('Pacific/Apia', 'Pacific/Apia', 'Apia', 'WS', 'Samoa (western)'),
    'Pacific/Auckland': ('Pacific/Auckland', 'Pacific/Auckland', 'Auckland', 'NZ', 'New Zealand'),
    'Pacific/Bougainville': ('Pacific/Bougainville', 'Pacific/Bougainville', 'Bougainville', 'PG', 'Papua New Guinea'),
    'Pacific/Chatham': ('Pacific/Chatham', 'Pacific/Chatham', 'Chatham', 'NZ', 'New Zealand'),
    'Pacific/Chuuk': ('Pacific/Chuuk', 'Pacific/Chuuk', 'Chuuk', 'FM', 'Micronesia'),
    'Pacific/Easter': ('Pacific/Easter', 'Pacific/Easter', 'Easter', 'CL', 'Chile'),
    'Pacific/Efate': ('Pacific/Efate', 'Pacific/Efate', 'Efate', 'VU', 'Vanuatu'),
    'Pacific/Fakaofo': ('Pacific/Fakaofo', 'Pacific/Fakaofo', 'Fakaofo', 'TK', 'Tokelau'),
    'Pacific/Fiji': ('Pacific/Fiji', 'Pacific/Fiji', 'Fiji', 'FJ', 'Fiji'),
    'Pacific/Funafuti': ('Pacific/Funafuti', 'Pacific/Funafuti', 'Funafuti', 'TV', 'Tuvalu'),
    'Pacific/Galapagos': ('Pacific/Galapagos', 'Pacific/Galapagos', 'Galapagos', 'EC', 'Ecuador'),
    'Pacific/Gambier': ('Pacific/Gambier', 'Pacific/Gambier', 'Gambier', 'PF', 'French Polynesia'),
    'Pacific/Guadalcanal': ('Pacific/Guadalcanal', 'Pacific/Guadalcanal', 'Guadalcanal', 'SB', 'Solomon Islands'),
    'Pacific/Guam': ('Pacific/Guam', 'Pacific/Guam', 'Guam', 'GU', 'Guam'),
    'Pacific/Honolulu': ('Pacific/Honolulu', 'Pacific/Honolulu', 'Honolulu', 'US', 'United States'),
    'Pacific/Kanton': ('Pacific/Kanton', 'Pacific/Kanton', 'Kanton', 'KI', 'Kiribati'),
    'Pacific/Kiritimati': ('Pacific/Kiritimati', 'Pacific/Kiritimati', 'Kiritimati', 'KI', 'Kiribati'),
    'Pacific/Kosrae': ('Pacific/Kosrae', 'Pacific/Kosrae', 'Kosrae', 'FM', 'Micronesia'),
    'Pacific/Kwajalein': ('Pacific/Kwajalein', 'Pacific/Kwajalein', 'Kwajalein', 'MH', 'Marshall Islands'),
    'Pacific/Majuro': ('Pacific/Majuro', 'Pacific/Majuro', 'Majuro', 'MH', 'Marshall Islands'),
    'Pacific/Marquesas': ('Pacific/Marquesas', 'Pacific/Marquesas', 'Marquesas', 'PF', 'French Polynesia'),
    'Pacific/Midway': ('Pacific/Midway', 'Pacific/Midway', 'Midway', 'UM', 'US minor outlying islands'),
    'Pacific/Nauru': ('Pacific/Nauru', 'Pacific/Nauru', 'Nauru', 'NR', 'Nauru'),
    'Pacific/Niue': ('Pacific/Niue', 'Pacific/Niue', 'Niue', 'NU', 'Niue'),
    'Pacific/Norfolk': ('Pacific/Norfolk', 'Pacific/Norfolk', 'Norfolk', 'NF', 'Norfolk Island'),
    'Pacific/Noumea': ('Pacific/Noumea', 'Pacific/Noumea', 'Noumea', 'NC', 'New Caledonia'),
    'Pacific/Pago_Pago': ('Pacific/Pago_Pago', 'Pacific/Pago Pago', 'Pago Pago', 'AS', 'Samoa (American)'),
    'Pacific/Palau': ('Pacific/Palau', 'Pacific/Palau', 'Palau', 'PW', 'Palau'),
    'Pacific/Pitcairn': ('Pacific/Pitcairn', 'Pacific/Pitcairn', 'Pitcairn', 'PN', 'Pitcairn'),
    'Pacific/Pohnpei': ('Pacific/Pohnpei', 'Pacific/Pohnpei', 'Pohnpei', 'FM', 'Micronesia'),
    'Pacific/Port_Moresby': ('Pacific/Port_Moresby', 'Pacific/Port Moresby', 'Port Moresby', 'PG', 'Papua New Guinea'),
    'Pacific/Rarotonga': ('Pacific/Rarotonga', 'Pacific/Rarotonga', 'Rarotonga', 'CK', 'Cook Islands'),
    'Pacific/Saipan': ('Pacific/Saipan', 'Pacific/Saipan', 'Saipan', 'MP', 'Northern Mariana Islands'),
    'Pacific/Tahiti': ('Pacific/Tahiti', 'Pacific/Tahiti', 'Tahiti', 'PF', 'French Polynesia'),
    'Pacific/Tarawa': ('Pacific/Tarawa', 'Pacific/Tarawa', 'Tarawa', 'KI', 'Kiribati'),
    'Pacific/Tongatapu': ('Pacific/Tongatapu', 'Pacific/Tongatapu', 'Tongatapu', 'TO', 'Tonga'),
    'Pacific/Wake': ('Pacific/Wake', 'Pacific/Wake', 'Wake', 'UM', 'US minor outlying islands'),
    'Pacific/Wallis': ('Pacific/Wallis', 'Pacific/Wallis', 'Wallis', 'WF', 'Wallis & Futuna'),
}


URI_SCHEMES = {
    'aaa',
    'aaas',
    'about',
    'acap',
    'acct',
    'acd',
    'acr',
    'adiumxtra',
    'adt',
    'afp',
    'afs',
    'aim',
    'amss',
    'android',
    'appdata',
    'apt',
    'ar',
    'ari',
    'ark',
    'at',
    'attachment',
    'aw',
    'barion',
    'bb',
    'beshare',
    'bitcoin',
    'bitcoincash',
    'blob',
    'bluetooth',
    'bolo',
    'brid',
    'browserext',
    'cabal',
    'calculator',
    'callto',
    'cap',
    'cast',
    'casts',
    'chrome',
    'chrome-extension',
    'cid',
    'coap',
    'coap+tcp',
    'coap+ws',
    'coaps',
    'coaps+tcp',
    'coaps+ws',
    'com-eventbrite-attendee',
    'content',
    'content-type',
    'crid',
    'cstr',
    'cvs',
    'dab',
    'dat',
    'data',
    'dav',
    'dhttp',
    'diaspora',
    'dict',
    'did',
    'dis',
    'dlna-playcontainer',
    'dlna-playsingle',
    'dns',
    'dntp',
    'doi',
    'dpp',
    'drm',
    'drop',
    'dtmi',
    'dtn',
    'dvb',
    'dvx',
    'dweb',
    'ed2k',
    'eid',
    'elsi',
    'embedded',
    'ens',
    'ethereum',
    'example',
    'facetime',
    'fax',
    'feed',
    'feedready',
    'fido',
    'file',
    'filesystem',
    'finger',
    'first-run-pen-experience',
    'fish',
    'fm',
    'ftp',
    'fuchsia-pkg',
    'geo',
    'gg',
    'git',
    'gitoid',
    'gizmoproject',
    'go',
    'gopher',
    'graph',
    'grd',
    'gtalk',
    'h323',
    'ham',
    'hcap',
    'hcp',
    'hs20',
    'http',
    'https',
    'hxxp',
    'hxxps',
    'hydrazone',
    'hyper',
    'iax',
    'icap',
    'icon',
    'ilstring',
    'im',
    'imap',
    'info',
    'iotdisco',
    'ipfs',
    'ipn',
    'ipns',
    'ipp',
    'ipps',
    'irc',
    'irc6',
    'ircs',
    'iris',
    'iris.beep',
    'iris.lwz',
    'iris.xpc',
    'iris.xpcs',
    'isostore',
    'itms',
    'jabber',
    'jar',
    'jms',
    'keyparc',
    'lastfm',
    'lbry',
    'ldap',
    'ldaps',
    'leaptofrogans',
    'lid',
    'lorawan',
    'lpa',
    'lvlt',
    'machineprovisioningprogressreporter',
    'magnet',
    'mailserver',
    'mailto',
    'maps',
    'market',
    'matrix',
    'message',
    'microsoft.windows.camera',
    'microsoft.windows.camera.multipicker',
    'microsoft.windows.camera.picker',
    'mid',
    'mms',
    'modem',
    'mongodb',
    'moz',
    'ms-access',
    'ms-appinstaller',
    'ms-browser-extension',
    'ms-calculator',
    'ms-drive-to',
    'ms-enrollment',
    'ms-excel',
    'ms-eyecontrolspeech',
    'ms-gamebarservices',
    'ms-gamingoverlay',
    'ms-getoffice',
    'ms-help',
    'ms-infopath',
    'ms-inputapp',
    'ms-launchremotedesktop',
    'ms-lockscreencomponent-config',
    'ms-media-stream-id',
    'ms-meetnow',
    'ms-mixedrealitycapture',
    'ms-mobileplans',
    'ms-newsandinterests',
    'ms-officeapp',
    'ms-people',
    'ms-personacard',
    'ms-project',
    'ms-powerpoint',
    'ms-publisher',
    'ms-recall',
    'ms-remotedesktop',
    'ms-remotedesktop-launch',
    'ms-restoretabcompanion',
    'ms-screenclip',
    'ms-screensketch',
    'ms-search',
    'ms-search-repair',
    'ms-secondary-screen-controller',
    'ms-secondary-screen-setup',
    'ms-settings',
    'ms-settings-airplanemode',
    'ms-settings-bluetooth',
    'ms-settings-camera',
    'ms-settings-cellular',
    'ms-settings-cloudstorage',
    'ms-settings-connectabledevices',
    'ms-settings-displays-topology',
    'ms-settings-emailandaccounts',
    'ms-settings-language',
    'ms-settings-location',
    'ms-settings-lock',
    'ms-settings-nfctransactions',
    'ms-settings-notifications',
    'ms-settings-power',
    'ms-settings-privacy',
    'ms-settings-proximity',
    'ms-settings-screenrotation',
    'ms-settings-wifi',
    'ms-settings-workplace',
    'ms-spd',
    'ms-stickers',
    'ms-sttoverlay',
    'ms-transit-to',
    'ms-useractivityset',
    'ms-virtualtouchpad',
    'ms-visio',
    'ms-walk-to',
    'ms-whiteboard',
    'ms-whiteboard-cmd',
    'ms-widgetboard',
    'ms-widgets',
    'ms-word',
    'msnim',
    'msrp',
    'msrps',
    'mss',
    'mt',
    'mtqp',
    'mumble',
    'mupdate',
    'mvn',
    'mvrp',
    'mvrps',
    'news',
    'nfs',
    'ni',
    'nih',
    'nntp',
    'notes',
    'num',
    'ocf',
    'oid',
    'onenote',
    'onenote-cmd',
    'opaquelocktoken',
    'openid',
    'openpgp4fpr',
    'otpauth',
    'p1',
    'pack',
    'palm',
    'paparazzi',
    'payment',
    'payto',
    'pkcs11',
    'platform',
    'pop',
    'pres',
    'prospero',
    'proxy',
    'pwid',
    'psyc',
    'pttp',
    'qb',
    'query',
    'quic-transport',
    'redis',
    'rediss',
    'reload',
    'res',
    'resource',
    'rmi',
    'rsync',
    'rtmfp',
    'rtmp',
    'rtsp',
    'rtsps',
    'rtspu',
    'sarif',
    'secondlife',
    'secret-token',
    'service',
    'session',
    'sftp',
    'sgn',
    'shc',
    'shelter',
    'shttp',
    'sieve',
    'simpleledger',
    'simplex',
    'sip',
    'sips',
    'skype',
    'smb',
    'smp',
    'sms',
    'smtp',
    'snews',
    'snmp',
    'soap.beep',
    'soap.beeps',
    'soldat',
    'spiffe',
    'spotify',
    'ssb',
    'ssh',
    'starknet',
    'steam',
    'stun',
    'stuns',
    'submit',
    'svn',
    'swh',
    'swid',
    'swidpath',
    'tag',
    'taler',
    'teamspeak',
    'teapot',
    'teapots',
    'tel',
    'teliaeid',
    'telnet',
    'tftp',
    'things',
    'thismessage',
    'thzp',
    'tip',
    'tn3270',
    'tool',
    'turn',
    'turns',
    'tv',
    'udp',
    'unreal',
    'upt',
    'urn',
    'ut2004',
    'uuid-in-package',
    'v-event',
    'vemmi',
    'ventrilo',
    'ves',
    'videotex',
    'vnc',
    'view-source',
    'vscode',
    'vscode-insiders',
    'vsls',
    'w3',
    'wais',
    'web3',
    'wcr',
    'webcal',
    'web+ap',
    'wifi',
    'wpid',
    'ws',
    'wss',
    'wtai',
    'wyciwyg',
    'xcon',
    'xcon-userid',
    'xfire',
    'xmlrpc.beep',
    'xmlrpc.beeps',
    'xmpp',
    'xftp',
    'xrcp',
    'xri',
    'ymsgr',
    'z39.50',
    'z39.50r',
    'z39.50s',
}



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Literal

from gajim.common import app
from gajim.common.const import SHOW_LIST
from gajim.common.const import SHOW_STRING
from gajim.common.const import SHOW_STRING_MNEMONIC


def get_client_status(account: str) -> str:
    client = app.get_client(account)
    if client.state.is_disconnected:
        return 'offline'

    if (
        client.state.is_reconnect_scheduled
        or client.state.is_connecting
        or client.state.is_connected
    ):
        return 'connecting'

    return client.status


def get_global_show() -> str:
    maxi = 0
    for client in app.get_clients():
        if not app.settings.get_account_setting(
            client.account, 'sync_with_global_status'
        ):
            continue
        status = get_client_status(client.account)
        index = SHOW_LIST.index(status)
        maxi = max(index, maxi)

    return SHOW_LIST[maxi]


def get_global_status_message() -> str:
    maxi = 0
    status_message = ''
    for client in app.get_clients():
        if not app.settings.get_account_setting(
            client.account, 'sync_with_global_status'
        ):
            continue
        index = SHOW_LIST.index(client.status)
        if index > maxi:
            maxi = index
            status_message = client.status_message
    return status_message


def statuses_unified() -> bool:
    '''
    Test if all statuses are the same
    '''
    reference = None
    for client in app.get_clients():
        account = client.account
        if not app.settings.get_account_setting(account, 'sync_with_global_status'):
            continue

        if reference is None:
            reference = get_client_status(account)

        elif reference != get_client_status(account):
            return False
    return True


def get_uf_show(show: str, use_mnemonic: bool = False) -> str:
    if use_mnemonic:
        return SHOW_STRING_MNEMONIC[show]
    return SHOW_STRING[show]


def get_idle_status_message(
    state: Literal['away'] | Literal['xa'], status_message: str
) -> str:
    if state == 'away':
        message = app.settings.get('autoaway_message')
        idle_time = app.settings.get('autoawaytime')
    else:
        message = app.settings.get('autoxa_message')
        idle_time = app.settings.get('autoxatime')

    if not message:
        return status_message

    message = message.replace('$S', '%(status)s')
    message = message.replace('$T', '%(time)s')

    return message % {'status': status_message, 'time': idle_time}
